; ==================================================================================================
; Title:   XTreeView.inc
; Author:  G. Friedrich
; Version: 1.0.0
; Purpose: Support of XTreeView objects for ObjAsm32.
; Notes:   Version 1.0.0, July 2007
;            - First release.
; ==================================================================================================

; Features:
; 
; Smooth scrolling.
; Customized drawing.
; Each node can have different heights.
; Fast drawing. Only nodes in view are drawn.
; Streaming is possible.
; No flicker due to backbuffering.
; High degree of customization due to inheritance.
; Full key navigation implemented.
; Mouse wheel support for both scrollers. The Shift key is used to scroll horizontally.
; Drag and drop is preformed with one of the mouse buttons.
;   The other button remains free to expand or collapse branches.
; Autoscrolling if we are in the "scroll activation area".
; Multiple node insertion modes implemented.
; Flexible intelligent cursor management.


; Some rules:
; 
; 1. The owner of all nodes is the XTreeView object.
; 2. The parent of a node is its hirarchical parent node in the tree.
; 3. Root nodes have NULL as parent.
; 4. Branch width and height are fields that contain the dims of a branch considdering the expanded
;    state of their childs. A node may have different values in the dHeight and dBranchHeight field.
;    You have to take the correct one depending on the node expansion state.
; 5. Non visible nodes should have dHeigt = dWidth = 0.
; 6. To use double clicks, enable CS_DBLCLKS in the class definition in the Startup BoundMethod of
;    the derived object.


; XTreeView public methods:
; 
; Init / Done
; Insert / Remove / Dispose
; Select / Draw


;Error codes
XTV_INVALID_POINTER           equ   2000000
XTV_UNKNOWN_INSERTION_MODE    equ   2000001


XTreeViewID                   equ   1000000
XTreeViewNodeID               equ   1000001

;XTreeViewNode States
XTVNS_DISABLED                equ   BIT00         ;Must be the first one
XTVNS_EXPANDED                equ   BIT01
XTVNS_HOTTRACK                equ   BIT02

XTVNS_SEL_PATH                equ   BIT08         ;Used to mark the path to the selected node
XTVNS_VIS_PATH                equ   BIT09         ;Used to mark the path to the first visible node


;XTreeView States
XTVS_DISABLED                 equ   BIT00         ;Must be the first one
XTVS_VISIBLE                  equ   BIT01         ;Visibility flag
XTVS_FOCUSED                  equ   BIT02
XTVS_LOCKED                   equ   BIT03         ;No visual update
XTVS_DRAGGINGSTARTED          equ   BIT04         ;A drag operation has been stared
XTVS_DRAGGING                 equ   BIT05         ;A drag operation is beeing performed
XTVS_DRAGGINGABORTED          equ   BIT06         ;The drag operation was aborted but LMB is still down

;XTreeView Flags
XTVF_RESOURCE                 equ   BIT01         ;Control was loaded from a resource
XTVF_SHOW_LINES               equ   BIT02
XTVF_SHOW_TOGGLER             equ   BIT03
XTVF_COLLAPSE_ON_DRAG         equ   BIT04
XTVF_EXPAND_PARENT_ON_DROP    equ   BIT05
XTVF_KEEP_COLLAPSED_ON_DRAG   equ   BIT06

XTV_FLAGS                     equ   XTVF_SHOW_LINES or XTVF_SHOW_TOGGLER or \
                                    XTVF_EXPAND_PARENT_ON_DROP or \
                                    XTVF_COLLAPSE_ON_DRAG or XTVF_KEEP_COLLAPSED_ON_DRAG

;Metrics
XTV_SCROLL_ACTIVATION         equ   15
XTV_TOGGLER_SIZE              equ   9


;Insertion modes
XTVN_INSERT_NOT_ALLOWED       equ   0
XTVN_INSERT_AS_FIRST_CHILD    equ   1
XTVN_INSERT_AS_LAST_CHILD     equ   2
XTVN_INSERT_AS_PREV_SIBLING   equ   3
XTVN_INSERT_AS_NEXT_SIBLING   equ   4


XTV_DEF struc
  dCtlID        dword     ?   ;Contol ID
  dStyle        dword     ?   ;Style
  dExStyle      dword     ?   ;Extended style
  pTitle        Pointer   ?   ;-> control title
  sdPosX        sdword    ?   ;X position
  sdPosY        sdword    ?   ;Y position
  dWidth        dword     ?   ;Width
  dHeight       dword     ?   ;Height
XTV_DEF ends



;  Node layout
;  
;
;
;  |<------ Node indent ----->|<----- Node width ----->|
;  |                          |                        |
;  |                          |                        |
;
;
;         sdNodeTop
;  |                                   |     /\
;  |                       |  Con-      |          |      
;  |                  [-]---|  tent  |  TextRect  |  |          |      Node height
;  |                Toggler |  Rect      |          |      
;  |                 Rect               |     \/
;  |                                     |      sdNodeBot
;  |                              |                      |     |     /\
;  |                          ----|      First Child     |     |      
;  |                              |                      |     |      
;  |                                     |      
;  |                                 |      
;  |                              |                          | |      
;  |                         [+]---|      Next Child          | |      Branch height
;  |                              |                          | |      
;  |                                 |      
;  |                                     |      
;  |                              |                      |     |      
;  |                           ----|      Last Child      |     |      
;  |                               |                      |     |      
;  |                                      |     \/
;        
;
;
;  |                                 |                          | |
;  |                                 |                          | |
;  |                                 |<----- Branch width ----->| |
;  |                                                              |
;  |<--------------------------- View width --------------------->|
;
;
;
;  Notes: - The ToggerRect vertical position is aligned with the ContentRect.
;         - The space between the parent vertical lines is defined by dNodeIndent.
;         - To add new node RECTs, XTVN_DISP_INFO must be expanded. In this case, the
;           new structure size must be updated in the dDispInfoSize data member.
;         - Branch metrics are computed in the current expansion state of child nodes.


XTVN_DISP_INFO struct
  pNode         Pointer   ?         ;-> node
  sdNodeTop     sdword    ?         ;Coord in hMemDC when drawn
  sdNodeBot     sdword    ?         ;Coord in hMemDC when drawn
  TogglerRect   RECT      {}        ;Coords in hMemDC when drawn
  ContentRect   RECT      {}        ;Coords in hMemDC when drawn
  TextRect      RECT      {}        ;Coords in hMemDC when drawn
XTVN_DISP_INFO ends


Object XTreeViewNode, XTreeViewNodeID, Streamable
  StaticMethod      CanDragMe,          dword
  StaticMethod      CanDropHere,        Pointer, dword, dword
  StaticMethod      ContextPopup,       dword, dword
  RedefineMethod    Done
  RedefineMethod    Init,               Pointer, dword, dword, dword
  StaticMethod      MouseHit,           dword, Pointer, Pointer, dword
  StaticMethod      MouseHitDbl,        dword, Pointer, Pointer, dword
  StaticMethod      Render,             Pointer                   ;-> XTVN_DISP_INFO
  StaticMethod      RenderContent,      Pointer
  StaticMethod      RenderText,         Pointer, Pointer
  StaticMethod      RenderTreeStruc,    Pointer
  StaticMethod      Toggle

  DefineVariable    pParent,            Pointer,      NULL        ;-> Parent node
  DefineVariable    pNextSibling,       Pointer,      NULL        ;-> Next sibling node
  DefineVariable    pPrevSibling,       Pointer,      NULL        ;-> Previous sibling node
  DefineVariable    dState,             dword,        0
  DefineVariable    dHeight,            dword,        0
  DefineVariable    dWidth,             dword,        0

  DefineVariable    pFirstChild,        Pointer,      NULL        ;-> First child node
  DefineVariable    pLastChild,         Pointer,      NULL        ;-> Last child node
  DefineVariable    dBranchHeight,      dword,        0
  DefineVariable    dBranchWidth,       dword,        0
ObjectEnd


Object XTreeView, XTreeViewID, WinPrimer
  StaticMethod      CanDropHere,        Pointer, dword, dword
  StaticMethod      Dispatch,           Handle, dword, dword, dword
  StaticMethod      Dispose,            Pointer
  RedefineMethod    Done
  StaticMethod      Draw
  StaticMethod      Drop,               Pointer, Pointer, dword
  StaticMethod      Find1stVisibleNode, dword
  StaticMethod      GetDispInfo,        Pointer
  RedefineMethod    Init,               Pointer, Handle, Pointer
  StaticMethod      InitRes,            Handle
  StaticMethod      Insert,             Pointer, Pointer, dword
  StaticMethod      MoveInViewBottom,   Pointer
  StaticMethod      MoveInViewTop,      Pointer
  StaticMethod      OnButtonDblClk,     dword, dword, dword
  StaticMethod      OnButtonDown,       dword, dword, dword
  StaticMethod      OnButtonUp,         dword, dword, dword
  StaticMethod      OnCreate,           dword, dword
  StaticMethod      OnDestroy,          dword, dword
  StaticMethod      OnEnable,           dword, dword
  StaticMethod      OnGetDlgCode,       dword, dword
  StaticMethod      OnGetFont,          dword, dword
  StaticMethod      OnHScroll,          dword, dword
  StaticMethod      OnKeyDown,          dword, dword
  StaticMethod      OnKeyUp,            dword, dword
  StaticMethod      OnKillFocus,        dword, dword
  StaticMethod      OnLButtonDblClk,    dword, dword
  StaticMethod      OnLButtonDown,      dword, dword
  StaticMethod      OnLButtonUp,        dword, dword
  StaticMethod      OnMButtonDblClk,    dword, dword
  StaticMethod      OnMButtonDown,      dword, dword
  StaticMethod      OnMButtonUp,        dword, dword
  StaticMethod      OnMouseMove,        dword, dword
  StaticMethod      OnMouseWheel,       dword, dword
  StaticMethod      OnNcDestroy,        dword, dword
  StaticMethod      OnNcHitTest,        dword, dword
  StaticMethod      OnRButtonDblClk,    dword, dword
  StaticMethod      OnRButtonDown,      dword, dword
  StaticMethod      OnRButtonUp,        dword, dword
  StaticMethod      OnPaint,            dword, dword
  StaticMethod      OnSetFocus,         dword, dword
  StaticMethod      OnSetFont,          dword, dword
  StaticMethod      OnSetRedraw,        dword, dword
  StaticMethod      OnSize,             dword, dword
  StaticMethod      OnVScroll,          dword, dword
  StaticMethod      Remove,             Pointer
  StaticMethod      RenderBackground,   Pointer                   ;-> RECT
  StaticMethod      Select,             Pointer
  StaticMethod      Set1stVisibleNode,  Pointer
  StaticMethod      SetDragCursor,      dword, dword
  BoundMethod       Startup
  StaticMethod      WndProc,            dword, dword, dword

  Event     OnKeyDown,        WM_KEYDOWN
  Event     OnKeyUp,          WM_KEYUP
  Event     OnMouseMove,      WM_MOUSEMOVE
  Event     OnMouseWheel,     WM_MOUSEWHEEL
  Event     OnNcHitTest,      WM_NCHITTEST
  Event     OnLButtonDblClk,  WM_LBUTTONDBLCLK
  Event     OnLButtonDown,    WM_LBUTTONDOWN
  Event     OnLButtonUp,      WM_LBUTTONUP
  Event     OnMButtonDblClk,  WM_MBUTTONDBLCLK
  Event     OnMButtonDown,    WM_MBUTTONDOWN
  Event     OnMButtonUp,      WM_MBUTTONUP
  Event     OnRButtonDblClk,  WM_RBUTTONDBLCLK
  Event     OnRButtonDown,    WM_RBUTTONDOWN
  Event     OnRButtonUp,      WM_RBUTTONUP
  Event     OnPaint,          WM_PAINT
  Event     OnSize,           WM_SIZE
  Event     OnHScroll,        WM_HSCROLL
  Event     OnVScroll,        WM_VSCROLL
  Event     OnKillFocus,      WM_KILLFOCUS
  Event     OnSetFocus,       WM_SETFOCUS
  Event     OnCreate,         WM_CREATE
  Event     OnDestroy,        WM_DESTROY
  Event     OnSetFont,        WM_SETFONT
  Event     OnGetFont,        WM_GETFONT
  Event     OnEnable,         WM_ENABLE
  Event     OnNcDestroy,      WM_NCDESTROY
  Event     OnGetDlgCode,     WM_GETDLGCODE
  Event     OnSetRedraw,      WM_SETREDRAW


  DefineVariable    dState,             dword,        XTVS_VISIBLE
  DefineVariable    dFlags,             dword,        XTV_FLAGS
  DefineVariable    pFirstVisibleNode,  Pointer,      NULL
  DefineVariable    pSelectedNode,      Pointer,      NULL
  DefineVariable    dFirstOffset,       dword,        0
  DefineVariable    hParent,            Handle,       0
  DefineVariable    ClientSize,         POINT,        {0,0}       ;Area excluding the scrollers
  DefineVariable    ClientOffset,       POINT,        {2,2}
  DefineVariable    TotalSize,          POINT,        {0,0}
  DefineVariable    TotalOffset,        POINT,        {0,0}
  DefineVariable    dTogglerSize,       dword,        XTV_TOGGLER_SIZE
  DefineVariable    dNodeIndent,        dword,        20          ;Fix size
  DefineVariable    hMemDC,             Handle,       0           ;Back DC
  DefineVariable    hBmpPrv,            Handle,       0
  DefineVariable    hBmpMem,            Handle,       0
  DefineVariable    dColorLine,         dword,        $RGB(192,192,192)
  DefineVariable    dColorToggler,      dword,        $RGB(0,0,0)
  DefineVariable    dColorBackground,   dword,        $RGB(255,255,255)
  DefineVariable    hFontText,          Handle,       0
  DefineVariable    hPenLine,           Handle,       0
  DefineVariable    hPenToggler,        Handle,       0
  DefineVariable    ArrowScrollValue,   POINT,        {1,1}
  DefineVariable    pFirstChild,        Pointer,      NULL        ;-> First child node
  DefineVariable    pLastChild,         Pointer,      NULL        ;-> Last child node
  DefineVariable    pHotTrackedNode,    Pointer,      NULL        ;-> Last hot tracked node
  DefineVariable    dDispInfoSize,      dword,        sizeof XTVN_DISP_INFO
  DefineVariable    hCursorArrow,       Handle,       0           ;Arrow
  DefineVariable    hCursorNo,          Handle,       0           ;Slashed circle
  DefineVariable    hCursorNode,        Handle,       0           ;Arrow with rectangle
  DefineVariable    hCursorFirst,       Handle,       0
  DefineVariable    hCursorLast,        Handle,       0
  DefineVariable    hCursorPrev,        Handle,       0
  DefineVariable    hCursorNext,        Handle,       0
  DefineVariable    dScrollActivation,  dword,        XTV_SCROLL_ACTIVATION
  DefineVariable    dDragMouseButton,   dword,        0


  Embed   DispInfoColl,   DwordCollection   ;XTVN_DISP_INFO structure collection of nodes in view
  Embed   DispInfoPool,   DataPool          ;Storage for the XTVN_DISP_INFO structures

ObjectEnd


DisplayInfoColl_DestroyItem  proto :Pointer, :Pointer     ;Make it public for derived objects


; ==================================================================================================
;     Macros
; ==================================================================================================

; 
; Macro:     XTV_BrowseBack
; Purpose:   Browses the TreeView back like pressing the up key.
; Arguments: Arg1: -> starting node. Must be a register!
; Return:    pNode -> previous node or NULL.
; Note:      No other registers are used.

XTV_BrowseBack macro pNode:req
    .if [pNode].XTreeViewNode.pPrevSibling == NULL
      mov pNode, [pNode].XTreeViewNode.pParent
    .else
      mov pNode, [pNode].XTreeViewNode.pPrevSibling
      .while XTV_IsExpanded(pNode)
        mov pNode, [pNode].XTreeViewNode.pLastChild
      .endw
    .endif
endm

; 
; Macro:     XTV_BrowseFrwd
; Purpose:   Browses the TreeView forward like pressing the down key.
; Arguments: Arg1: -> starting node. Must be a register!
; Return:    pNode -> next node or NULL.
; Note:      No other registers are used.

XTV_BrowseFrwd macro pNode:req
    .if XTV_IsExpanded(pNode)
      mov pNode, [pNode].XTreeViewNode.pFirstChild
    .else
      .repeat
        .if [pNode].XTreeViewNode.pNextSibling == NULL
          mov pNode, [pNode].XTreeViewNode.pParent
        .else
          mov pNode, [pNode].XTreeViewNode.pNextSibling
          .break
        .endif
      .until pNode == NULL
    .endif
endm

; 
; Macro:     XTV_SimulateMouseMove
; Purpose:   Simulates a mouse movement.
; Arguments: Arg1: initial value for the return value.
;            Arg2: List of keys to be tested for.

XTV_SimulateMouseMove macro InitialValue:req, Args:vararg
    ifidn <InitialValue>, <0>
      xor ebx, ebx
    else
      mov ebx, MK_&InitialValue&
    endif
    for Arg, <Args>
      .if $invoke(GetAsyncKeyState, VK_&Arg&) != 0
        or ebx, MK_&Arg&
      .endif
    endm
    invoke GetCursorPos, addr CursorPos
    invoke ScreenToClient, [esi].hWnd, addr CursorPos
    mov eax, CursorPos.y
    shl eax, 16
    mov ecx, CursorPos.x
    mov ax, cx
    OCall esi.OnMouseMove, ebx, eax
    xor eax, eax                                        ;;Method return value
    ExitMethod                                          ;;Exit the metod here
endm

; 
; Macro:     XTV_GetBottomViewNode
; Purpose:   Returns the node at the bottom of the view. If we don't reach the bottom of the view,
;            it returns NULL in the return register and in the optional register the last node.
; Arguments: Arg1: XTreeView regiter pointer.
;            Arg2: return register.
;            Arg3: auxiliar register. Will be trashed.
;            Arg4: (optional) Last node.
; Return:    return register -> bottom node or NULL.

XTV_GetBottomViewNode macro pXTVReg:req, RetReg:req, AuxReg:req, LastNode:=<>
    mov RetReg, [pXTVReg].pFirstVisibleNode
    mov AuxReg, [RetReg].XTreeViewNode.dHeight
    sub AuxReg, [pXTVReg].dFirstOffset
    .while AuxReg <= [pXTVReg].ClientSize.y
      ifdef LastNode
        mov LastNode, RetReg
      endif
      XTV_BrowseFrwd RetReg
      .break .if RetReg == NULL
      add AuxReg, [RetReg].XTreeViewNode.dHeight
    .endw
endm

; 
; Macro:     XTV_GetBranchMaxWidth
; Purpose:   Returns the width of a tree branch
; Arguments: Arg1: first node in branch.
;            Arg2: -> XTreeView.
; Return:    ebx = width.

XTV_GetBranchMaxWidth macro pStartNode:req, pXTV:req    ;;-> first node in branch,-> XTreeView
    mov eax, pStartNode
    xor ebx, ebx                                        ;;ebx = max width
    .while eax != NULL
      ;Calc the max width
      mov ecx, [eax].XTreeViewNode.dWidth
      Maxi ebx, ecx, edx                                ;;Trash ecx and edx
      .if XTV_IsExpanded(eax)
        mov ecx, [eax].XTreeViewNode.dBranchWidth
        if (opattr(pXTV)) and 00010000b                 ;;Is pXTV a register?
          add ecx, [pXTV].dNodeIndent
        else
          mov edx, pXTV
          add ecx, [edx].XTreeView.dNodeIndent
        endif
        Maxi ebx, ecx, edx                              ;;Trash ecx and edx
      .endif
      mov eax, [eax].XTreeViewNode.pNextSibling
    .endw
endm

; 
; Macro:     XTV_IsCollapsed
; Purpose:   Expression to check if a node is collapsed.
; Arguments: Arg1: -> Node. Must be a register!
; Return:    Expression to be used in an .if directive.

XTV_IsCollapsed macro pNode:req
    exitm <([pNode].XTreeViewNode.pFirstChild !!= NULL) && \
           !!([pNode].XTreeViewNode.dState & XTVNS_EXPANDED)>
endm

; 
; Macro:     XTV_IsExpanded
; Purpose:   Expression to check if a node is expanded.
; Arguments: Arg1: -> Node. Must be a register!
; Return:    Expression to be used in an .if directive.

XTV_IsExpanded macro pNode:req
    exitm <([pNode].XTreeViewNode.pFirstChild !!= NULL) && \
           ([pNode].XTreeViewNode.dState & XTVNS_EXPANDED)>
endm

; 
; Macro:     XTV_IsPointInRect
; Purpose:   Expression to check if a point is contained in a rectangle.
; Arguments: Arg1: X coord.
;            Arg2: Y coord.
;            Arg3: RECT struct.
; Return:    Expression to be used in an .if directive.

XTV_IsPointInRect macro XReg:req, YReg:req, Rct:req
    exitm <(sdword ptr YReg !>= Rct.top) && \           ;;Check first for y coords
           (sdword ptr YReg !< Rct.bottom) && \
           (sdword ptr XReg !>= Rct.left) && \          ;;and then for x coords
           (sdword ptr XReg !< Rct.right)>
endm


; ==================================================================================================

if IMPLEMENT

CStr szXTV_Class, "XTreeView"

.data
XTV_LogFontText LOGFONT <0, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, DEFAULT_CHARSET, \
                         OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, CLEARTYPE_QUALITY, \
                         DEFAULT_PITCH or FF_DONTCARE, "Arial">


.code
; ==================================================================================================
;     XTreeViewNode
; ==================================================================================================

;##\

; 
; Method:    XTreeViewNode.CanDragMe
; Purpose:   Indicates if a drag operation can be started.
; Arguments: Arg1: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    Always FALSE => no dragging allowed.

Method XTreeViewNode.CanDragMe, NOFRAME, dKeyFlags:dword
    xor eax, eax
MethodEnd 8, eax

; 
; Method:    XTreeViewNode.CanDropHere
; Purpose:   Informs how the node on arg1 can be inserted when dropped. Standard behaviour.
; Arguments: Arg1: -> Node to be dropped.
;            Arg2: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
;            Arg3: Cursor position.
; Return:    0 = drop not allowed (XTVN_INSERT_NOT_ALLOWED).
;            1 = drop as first child of reference node (XTVN_INSERT_AS_FIRST_CHILD).
;            2 = drop as last child of reference node (XTVN_INSERT_AS_LAST_CHILD).
;            3 = drop before reference sibling node (XTVN_INSERT_AS_PREV_SIBLING).
;            4 = drop behind reference sibling node (XTVN_INSERT_AS_NEXT_SIBLING).

Method XTreeViewNode.CanDropHere, uses esi, pNode:Pointer, dKeyFlags:dword, dCursorPos:dword
    local CursorPos:POINT

    SetObject esi
    ;Verify if this object isn't a child of pNode
    mov ecx, esi
    .repeat
      mov ecx, [ecx].XTreeViewNode.pParent
    .until (ecx == NULL) || (ecx == pNode)

    .if ecx == pNode
      mov eax, XTVN_INSERT_NOT_ALLOWED
    .else
      PntS2Pnt CursorPos, dCursorPos
      OCall [esi].pOwner::XTreeView.GetDispInfo, esi
      .if eax != NULL
        mov ecx, [eax].XTVN_DISP_INFO.sdNodeTop
        add ecx, [eax].XTVN_DISP_INFO.sdNodeBot
        shr ecx, 1
        mov eax, pNode
        .ifBitSet dKeyFlags, MK_CONTROL
          .if sdword ptr ecx > CursorPos.y
            .if esi != [eax].XTreeViewNode.pNextSibling
              mov eax, XTVN_INSERT_AS_PREV_SIBLING
            .elseif esi != [eax].XTreeViewNode.pPrevSibling
              mov eax, XTVN_INSERT_AS_NEXT_SIBLING
            .else
              mov eax, XTVN_INSERT_NOT_ALLOWED
            .endif
          .else
            .if esi != [eax].XTreeViewNode.pPrevSibling
              mov eax, XTVN_INSERT_AS_NEXT_SIBLING
            .elseif esi != [eax].XTreeViewNode.pNextSibling
              mov eax, XTVN_INSERT_AS_PREV_SIBLING
            .else
              mov eax, XTVN_INSERT_NOT_ALLOWED
            .endif
          .endif
        .else
          .if sdword ptr ecx > CursorPos.y
            .if eax != [esi].pFirstChild
              mov eax, XTVN_INSERT_AS_FIRST_CHILD
            .elseif eax != [esi].pLastChild
              mov eax, XTVN_INSERT_AS_LAST_CHILD
            .else
              mov eax, XTVN_INSERT_NOT_ALLOWED
            .endif
          .else
            .if [esi].pLastChild == NULL
              mov eax, XTVN_INSERT_AS_FIRST_CHILD
            .elseif eax != [esi].pLastChild
              mov eax, XTVN_INSERT_AS_LAST_CHILD
            .elseif eax != [esi].pFirstChild
              mov eax, XTVN_INSERT_AS_FIRST_CHILD
            .else
              mov eax, XTVN_INSERT_NOT_ALLOWED
            .endif
          .endif
        .endif
      .endif
    .endif
MethodEnd, eax

; 
; Method:    XTreeViewNode.ContextPopup
; Purpose:   Placeholder for a context popup menu.
; Arguments: Arg1: x coord in screen coords.
;            Arg2: y coord in screen coords.
; Return:    Nothing.

Method XTreeViewNode.ContextPopup, NOFRAME, dXPos:dword, dYPos:dword
MethodEnd 12, NONE

; 
; Method:    XTreeViewNode.Done
; Purpose:   Finalizes the XTreeViewNode object and all childs in its branch.
; Arguments: None.
; Return:    Nothing.

Method XTreeViewNode.Done, uses esi
    ;DbgText "XTreeViewNode.Done"
    SetObject esi
    ;Destroy all nodes
    mov eax, [esi].pFirstChild
    .while eax != NULL
      push [eax].XTreeViewNode.pNextSibling
      Kill eax
      pop eax
    .endw
    ACall esi.Done
MethodEnd 4, NONE

; 
; Method:    XTreeViewNode.Init
; Purpose:   Initializes a XTreeViewNode object.
; Arguments: Arg1: -> owner object, XTreeView.
;            Arg2: Node width.
;            Arg3: Node height.
;            Arg4: Node state (XTVNS_DISABLED, XTVNS_EXPANDED, etc.).
; Return:    Nothing.

Method XTreeViewNode.Init, uses esi, pOwner:Pointer, dWidth:dword, dHeight:dword, dState:dword
    SetObject esi
    ACall esi.Init, pOwner
    m2m [esi].dWidth, dWidth, eax
    m2m [esi].dHeight, dHeight, eax
    m2m [esi].dState, dState, eax
MethodEnd, NONE

; 
; Method:    XTreeViewNode.MouseHit
; Purpose:   Checks if the mouse has clicked a hotspot and performs an action.
; Arguments: Arg1: button that triggered this action.
;            Arg2: -> clicked point.
;            Arg3: -> XTVN_DISP_INFO.
;            Arg4: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    eax = TRUE if and action was triggered, otherwise FALSE.

Method XTreeViewNode.MouseHit, uses ebx esi, dButton:dword, pPoint:Pointer, pDispInfo:Pointer, \
                                             dKeyFlags:dword
    local CRect:RECT

    ;DbgText "XTreeViewNode.MouseHit"
    .if pDispInfo != NULL
      SetObject esi
      mov ebx, [esi].pOwner

      ;Check for text hit
      mov ecx, pPoint
      mov edx, pDispInfo
      assume ebx:ptr XTreeView, edx:ptr XTVN_DISP_INFO
      mov eax, [ecx].POINT.x
      mov ecx, [ecx].POINT.y
      .if XTV_IsPointInRect(eax, ecx, [edx].TextRect) && !([ebx].dState & XTVS_DRAGGING)
        ;DbgText "Hit on Text"
        .if esi != [ebx].pSelectedNode
          OCall ebx::XTreeView.GetDispInfo, [ebx].pSelectedNode
          push [ebx].pSelectedNode
          push eax
          OCall ebx::XTreeView.Select, esi
          OCall esi.RenderContent, pDispInfo
          pop eax
          pop ecx
          .if eax != NULL
            OCall ecx::XTreeViewNode.RenderContent, eax
          .endif
        .endif

        .if !([ebx].dState & \
              (XTVS_DRAGGING or XTVS_DRAGGINGABORTED or XTVS_DRAGGINGSTARTED))
          .if $OCall(esi.CanDragMe, dKeyFlags)
            BitSet [ebx].dState, XTVS_DRAGGINGSTARTED
            mov eax, dKeyFlags
            and eax, (MK_LBUTTON or MK_MBUTTON or MK_RBUTTON)
            mov [ebx].dDragMouseButton, eax
          .endif
        .endif
        invoke UpdateWindow, [ebx].hWnd
        mov eax, TRUE

      ;Check for toggler hit
      .elseif XTV_IsPointInRect(eax, ecx, [edx].TogglerRect) && \
              !(([ebx].dFlags & XTVF_KEEP_COLLAPSED_ON_DRAG) && \
              (([ebx].dState & XTVS_DRAGGING) && (esi == [ebx].pSelectedNode)))
        ;DbgText "Hit on Toggler"
        OCall esi::XTreeViewNode.Toggle

        .if XTV_IsExpanded(esi)

          ;Correct the client area coords
          mov eax, [ebx].TotalSize.x
          .if eax > [ebx].ClientSize.x
            invoke ShowScrollBar, [ebx].hWnd, SB_HORZ, TRUE
            invoke GetClientRect, [ebx].hWnd, addr CRect
            m2m [ebx].ClientSize.x, CRect.right
            m2m [ebx].ClientSize.y, CRect.bottom
          .endif

          ;Node is now expanded; maybe we have to reposition the tree
          OCall ebx::XTreeView.GetDispInfo, esi
          mov eax, [eax].XTVN_DISP_INFO.sdNodeTop
          add eax, [ebx].TotalOffset.y

          mov edx, eax
          add edx, [esi].dHeight
          add edx, [esi].dBranchHeight
          sub edx, [ebx].ClientSize.y
          sub edx, [ebx].TotalOffset.y
          .if !Sign?
            mov ecx, eax                                  ;The last branch node is off view
            sub ecx, [ebx].TotalOffset.y
            sub ecx, edx
            .if !Sign?
              add eax, [esi].dHeight
              add eax, [esi].dBranchHeight
              sub eax, [ebx].ClientSize.y
            .endif
            OCall ebx::XTreeView.Find1stVisibleNode, eax
          .endif

        .else
          ;Node is now collapsed; we check if we have a hole at the bottom
          XTV_GetBottomViewNode ebx, ecx, eax, edx
          .if ecx == NULL
            OCall ebx::XTreeView.MoveInViewBottom, edx
          .endif
        .endif
        OCall ebx::XTreeView.Draw
        invoke UpdateWindow, [ebx].hWnd
        mov eax, TRUE

      .else
        xor eax, eax

      .endif
      assume edx:NOTHING, ebx:NOTHING
    .else
      xor eax, eax
    .endif
MethodEnd, eax

; 
; Method:    XTreeViewNode.MouseHitDbl
; Purpose:   Checks if the mouse has double clicked a hotspot and performs an action.
; Arguments: Arg1: button that triggered this action.
;            Arg2: -> clicked point.
;            Arg3: -> XTVN_DISP_INFO.
;            Arg4: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    eax = TRUE if and action was triggered, otherwise FALSE.

Method XTreeViewNode.MouseHitDbl, NOFRAME, dButton:dword, pPoint:Pointer, pDispInfo:Pointer, \
                                           dKeyFlags:dword
    xor eax, eax
MethodEnd 20, eax

; 
; Method:    XTreeViewNode.Render
; Purpose:   Draws the complete node.
; Arguments: Arg1: -> XTVN_DISP_INFO.
; Return:    Nothing.

Method XTreeViewNode.Render, uses esi, pDispInfo:Pointer
    ;DbgText "XTreeViewNode.Render"
    SetObject esi
    .if ([esi].dHeight != 0) || ([esi].dWidth != 0)     ;Don't draw nodes with height and width = 0
      mov edx, [esi].pOwner

      ;Calc the total indentation of this node
      mov ecx, [esi].pParent
      mov eax, [edx].XTreeView.ClientOffset.x
      mov edx, [esi].pOwner
      sub eax, [edx].XTreeView.TotalOffset.x
      add eax, [edx].XTreeView.dNodeIndent


      .while ecx != NULL
        mov ecx, [ecx].XTreeViewNode.pParent
        add eax, [edx].XTreeView.dNodeIndent
      .endw

      ;Fill XTVN_DISP_INFO.ContentRect info
      mov ecx, pDispInfo
      mov [ecx].XTVN_DISP_INFO.ContentRect.left, eax
      add eax, [esi].dWidth
      mov [ecx].XTVN_DISP_INFO.ContentRect.right, eax
      m2m [ecx].XTVN_DISP_INFO.ContentRect.top, [ecx].XTVN_DISP_INFO.sdNodeTop, eax
      m2m [ecx].XTVN_DISP_INFO.ContentRect.bottom, [ecx].XTVN_DISP_INFO.sdNodeBot, eax

      if DEBUGGING
      .if sdword ptr [ecx].XTVN_DISP_INFO.ContentRect.bottom < 0
        DbgWarning "Error while positioning the view"
      .endif
      endif

      ;Render the tree structure (lines and toggler)
      OCall esi.RenderTreeStruc, pDispInfo

      ;Render the content of the node
      OCall esi.RenderContent, pDispInfo

    .endif
MethodEnd, NONE

; 
; Method:    XTreeViewNode.RenderContent
; Purpose:   Draws the content of the node.
; Arguments: Arg1: -> XTVN_DISP_INFO structure. If NULL, no action is performed.
; Return:    Nothing.
; Note:      This method must update the coords of all RECT structures in the Content and must
;            invalidate the modified region.

Method XTreeViewNode.RenderContent, uses ebx edi esi, pDispInfo:Pointer
    local hPrevBrush:dword, InvRect:RECT

    ;DbgText "XTreeViewNode.RenderContent"
    mov edi, pDispInfo
    .if edi != NULL
      SetObject esi
      mov ebx, [esi].pOwner
      assume edi:ptr XTVN_DISP_INFO, ebx:ptr XTreeView

      ;Get the drawing rect
      m2m InvRect.left, [edi].ContentRect.left, eax
      m2m InvRect.top, [edi].sdNodeTop, eax
      m2m InvRect.right, [ebx].ClientSize.x, eax            ;Invalidate up to the end of the view
      m2m InvRect.bottom, [edi].sdNodeBot, eax

      ;Clear the background
      m2m [edi].TextRect.left, [edi].ContentRect.left, eax
      m2m [edi].TextRect.right, [edi].ContentRect.right, eax
      m2m [edi].TextRect.top, [edi].ContentRect.top, eax
      m2m [edi].TextRect.bottom, [edi].ContentRect.bottom, eax

      OCall ebx::XTreeView.RenderBackground, addr InvRect

      ;Draw a hatched rectangle to visualize the ContentRect
      .if esi == [ebx].pSelectedNode
        invoke CreateHatchBrush, HS_DIAGCROSS, $RGB(0,0,255)
      .else
        invoke CreateHatchBrush, HS_BDIAGONAL, $RGB(0,0,0)
      .endif
      mov hPrevBrush, $invoke(SelectObject, [ebx].hMemDC, eax)
      invoke Rectangle, [ebx].hMemDC, \
               [edi].ContentRect.left, [edi].ContentRect.top, \
               [edi].ContentRect.right, [edi].ContentRect.bottom
      invoke DeleteObject, $invoke(SelectObject, [ebx].hMemDC, hPrevBrush)

      ;Invalidate the new drawn rect
      invoke InvalidateRect, [ebx].hWnd, addr InvRect, FALSE
      assume edi:NOTHING, ebx:NOTHING
    .endif
MethodEnd, NONE

; 
; Method:    XTreeViewNode.RenderText
; Purpose:   Draws the content of the node.
; Arguments: Arg1: -> XTVN_DISP_INFO. If NULL, no action is performed.
;            Arg2: -> Text.
; Return:    Nothing.
; Note:      This method centers the TextRect verticaly to ContentRect.
;            The TextRect horizontal position is not changed.

Method XTreeViewNode.RenderText, uses ebx edi esi, pDispInfo:Pointer, pText:Pointer
    ;DbgText "XTreeViewNode.RenderText"
    mov ebx, pDispInfo
    .if ebx != NULL
      SetObject esi
      mov edi, [esi].pOwner
      assume ebx:ptr XTVN_DISP_INFO, edi:ptr XTreeView
      mov [ebx].TextRect.top, 0

      invoke DrawText, [edi].hMemDC, pText , -1, addr [ebx].TextRect, \
                        DT_SINGLELINE or DT_LEFT or DT_CALCRECT

      mov ecx, [ebx].ContentRect.bottom
      sub ecx, [ebx].ContentRect.top
      sub ecx, eax                                        ;eax = text height
      shr ecx, 1
      add ecx, [ebx].ContentRect.top
      mov [ebx].TextRect.top, ecx
      add ecx, eax
      mov [ebx].TextRect.bottom, ecx
      push [ebx].TextRect.right                           ;Save right coord

      ;Set max right size to draw the elipsis
      m2m [ebx].TextRect.right, [edi].ClientSize.x, eax
      invoke DrawText, [edi].hMemDC, pText , -1, addr [ebx].TextRect, \
                       DT_SINGLELINE or DT_LEFT or DT_VCENTER or DT_END_ELLIPSIS

      pop [ebx].TextRect.right                            ;Restore right coord
      assume ebx:NOTHING, edi:NOTHING
    .endif
MethodEnd, NONE

; 
; Method:    XTreeViewNode.RenderTreeStruc
; Purpose:   Draws the connection lines and toggler boxes ([+] or [-]).
; Arguments: Arg1: -> XTVN_DISP_INFO. If NULL, no action is performed.
; Return:    Nothing.
; Note:      This method updates the coords of the TogglerRect structure.

Method XTreeViewNode.RenderTreeStruc, uses ebx edi esi, pDispInfo:Pointer
    local Center:POINT, hPenPrev:Handle, InvRect:RECT

    mov ebx, pDispInfo
    .if ebx != NULL
      SetObject esi
      mov edi, [esi].pOwner
      assume ebx:ptr XTVN_DISP_INFO, edi:ptr XTreeView

      ;Get the drawing rect
      mov InvRect.left, 0
      m2m InvRect.top, [ebx].sdNodeTop, eax
      m2m InvRect.right, [ebx].ContentRect.left, eax
      m2m InvRect.bottom, [ebx].sdNodeBot, eax

      ;Clear background
      OCall edi::XTreeView.RenderBackground, addr InvRect

      ;Calc the center of the toggler box. It is vertically aligned with the ContentRect
      mov eax, [ebx].ContentRect.bottom
      sub eax, [ebx].ContentRect.top
      shr eax, 1
      add eax, [ebx].ContentRect.top
      mov Center.y, eax
      mov ecx, [ebx].ContentRect.left
      mov edx, [edi].dNodeIndent
      shr edx, 1
      sbb ecx, edx
      mov Center.x, ecx

      .if ([esi].pFirstChild == NULL) || !([edi].dFlags & XTVF_SHOW_TOGGLER)  ;No child nodes?
        .ifBitSet [edi].dFlags, XTVF_SHOW_LINES
          ;Check if the node is the top parent node
          mov hPenPrev, $invoke(SelectObject, [edi].hMemDC, [edi].hPenLine)
          .if [esi].pParent == NULL && [esi].pPrevSibling == NULL
            mov ecx, Center.y
          .else
            mov ecx, [ebx].sdNodeTop
          .endif
          ;Draw the connection lines
          invoke MoveToEx, [edi].hMemDC, Center.x, ecx, NULL
          .if [esi].pNextSibling == NULL                ;Is it an ending node?
            ;Last node
            invoke LineTo, [edi].hMemDC, Center.x, Center.y
          .else
            ;Any other node
            invoke LineTo, [edi].hMemDC, Center.x, [ebx].sdNodeBot
            invoke MoveToEx, [edi].hMemDC, Center.x, Center.y, NULL
          .endif
          invoke LineTo, [edi].hMemDC, [ebx].ContentRect.left, Center.y
          invoke MemZero, addr [ebx].TogglerRect, sizeof RECT
          invoke SelectObject, [edi].hMemDC, hPenPrev
        .endif
      .else
        ;Calc the rect of the toggler
        mov ecx, [ebx].ContentRect.left
        mov eax, [edi].dTogglerSize
        add eax, [edi].dNodeIndent
        shr eax, 1
        sub ecx, eax
        mov [ebx].TogglerRect.left, ecx
        add ecx, [edi].dTogglerSize
        mov [ebx].TogglerRect.right, ecx

        mov eax, [ebx].ContentRect.bottom
        sub eax, [ebx].ContentRect.top
        sub eax, [edi].dTogglerSize
        shr eax, 1
        mov ecx, [ebx].ContentRect.bottom
        sub ecx, eax
        mov [ebx].TogglerRect.bottom, ecx
        sub ecx, [edi].dTogglerSize
        mov [ebx].TogglerRect.top, ecx

        ;Draw the toggler rectangle
        mov hPenPrev, $invoke(SelectObject, [edi].hMemDC, [edi].hPenToggler)
        invoke Rectangle, [edi].hMemDC, \
                [ebx].TogglerRect.left, [ebx].TogglerRect.top, \
                [ebx].TogglerRect.right, [ebx].TogglerRect.bottom

        ;Draw the content of the toggler [+] or [-]
        mov eax, [ebx].TogglerRect.left
        add eax, 2
        ;Draw horizontal line of a minus sign
        invoke MoveToEx, [edi].hMemDC, eax, Center.y, NULL
        mov eax, [ebx].TogglerRect.right
        sub eax, 2
        invoke LineTo, [edi].hMemDC, eax, Center.y
        .ifBitClr [esi].dState, XTVNS_EXPANDED
          ;Make a plus
          mov eax, [ebx].TogglerRect.top
          add eax, 2
          invoke MoveToEx, [edi].hMemDC, Center.x, eax, NULL
          mov eax, [ebx].TogglerRect.bottom
          sub eax, 2
          invoke LineTo, [edi].hMemDC, Center.x, eax
        .endif
        invoke SelectObject, [edi].hMemDC, hPenPrev

        ;Draw the toggler connection lines
        .ifBitSet [edi].dFlags, XTVF_SHOW_LINES
          mov hPenPrev, $invoke(SelectObject, [edi].hMemDC, [edi].hPenLine)
          ;Check if the node is the top parent node
          .if [esi].pParent == NULL && [esi].pPrevSibling == NULL
            mov ecx, [ebx].TogglerRect.top
          .else
            mov ecx, [ebx].sdNodeTop
          .endif
          invoke MoveToEx, [edi].hMemDC, Center.x, ecx, NULL
          invoke LineTo, [edi].hMemDC, Center.x, [ebx].TogglerRect.top
          invoke MoveToEx, [edi].hMemDC, [ebx].TogglerRect.right, Center.y, NULL
          invoke LineTo, [edi].hMemDC, [ebx].ContentRect.left, Center.y
          .if [esi].pNextSibling != NULL                ;Is it not an ending node?
            invoke MoveToEx, [edi].hMemDC, Center.x, [ebx].TogglerRect.bottom, NULL
            invoke LineTo, [edi].hMemDC, Center.x, [ebx].sdNodeBot
          .endif
          invoke SelectObject, [edi].hMemDC, hPenPrev
        .endif
      .endif

      ;Draw vertical parent connection lines
      .ifBitSet [edi].dFlags, XTVF_SHOW_LINES
        mov hPenPrev, $invoke(SelectObject, [edi].hMemDC, [edi].hPenLine)
        .while TRUE

          mov esi, [esi].pParent
          .break .if esi == NULL
          mov eax, [edi].dNodeIndent
          sub Center.x, eax
          .if [esi].pNextSibling != NULL
            invoke MoveToEx, [edi].hMemDC, Center.x, [ebx].sdNodeTop, NULL
            invoke LineTo, [edi].hMemDC, Center.x, [ebx].sdNodeBot
          .endif
        .endw
        invoke SelectObject, [edi].hMemDC, hPenPrev
      .endif

      ;Invalidate the redrawn rect
      invoke InvalidateRect, [edi].hWnd, addr InvRect, FALSE

      assume ebx:NOTHING, edi:NOTHING
    .endif
MethodEnd, NONE

; 
; Method:    XTreeViewNode.Toggle
; Purpose:   Expands or collapses this node.
; Arguments: None.
; Return:    Nothing.

Method XTreeViewNode.Toggle, uses ebx edi esi
    ;DbgText "XTreeViewNode.Toggle"
    SetObject esi
    .ifBitSet [esi].dState, XTVNS_EXPANDED
      ;Collapse the branch

      BitClr [esi].dState, XTVNS_EXPANDED               ;Reverse it

      ;Manage node selection: if a leave is selected, select this node
      .ifBitSet [esi].dState, XTVNS_SEL_PATH
        mov eax, [esi].pOwner
        .ifBitClr [eax].XTreeView.dState, XTVS_DRAGGING
          OCall [esi].pOwner::XTreeView.Select, esi
        .endif
      .endif

      ;Calc new branch values in the modified path
      mov edi, [esi].pParent
      .while (edi != NULL) && ([edi].XTreeViewNode.dState & XTVNS_EXPANDED)

        mov eax, [esi].dBranchHeight
        sub [edi].XTreeViewNode.dBranchHeight, eax

        XTV_GetBranchMaxWidth [edi].XTreeViewNode.pFirstChild, [esi].pOwner
        mov [edi].XTreeViewNode.dBranchWidth, ebx

        mov edi, [edi].XTreeViewNode.pParent            ;Move one level up
      .endw

      .if edi == NULL                                   ;Did we reach the root node?
        mov eax, [esi].dBranchHeight
        mov edi, [esi].pOwner
        sub [edi].XTreeView.TotalSize.y, eax

        XTV_GetBranchMaxWidth [edi].XTreeView.pFirstChild, [esi].pOwner
        add ebx, [edi].XTreeView.dNodeIndent
        add ebx, [edi].XTreeView.ClientOffset.x
        mov [edi].XTreeView.TotalSize.x, ebx
      .endif

    .else
      ;Expand the branch
      BitSet [esi].dState, XTVNS_EXPANDED               ;Reverse it

      ;Calc new sizes
      mov edi, [esi].pParent
      .while (edi != NULL) && ([edi].XTreeViewNode.dState & XTVNS_EXPANDED)
        mov eax, [esi].dBranchHeight
        add [edi].XTreeViewNode.dBranchHeight, eax

        XTV_GetBranchMaxWidth [edi].XTreeViewNode.pFirstChild, [esi].pOwner
        mov [edi].XTreeViewNode.dBranchWidth, ebx

        mov edi, [edi].XTreeViewNode.pParent
      .endw
      .if edi == NULL                                   ;Did we reach the root node?
        mov eax, [esi].dBranchHeight
        mov edi, [esi].pOwner
        add [edi].XTreeView.TotalSize.y, eax

        XTV_GetBranchMaxWidth [edi].XTreeView.pFirstChild, [esi].pOwner
        add ebx, [edi].XTreeView.dNodeIndent
        add ebx, [edi].XTreeView.ClientOffset.x
        mov [edi].XTreeView.TotalSize.x, ebx
      .endif

    .endif

MethodEnd, NONE

;##/


; ==================================================================================================
;     XTreeView
; ==================================================================================================

;##\

; 
; Procedure: XTV_IsPointInContent
; Purpose:   Checks if a point is contained in the ContentRect.
; Arguments: Arg1: -> XTVN_DISP_INFO structure.
;            Arg2: x position in client coords.
;            Arg3: y position in client coords.
; Return:    TRUE / FALSE.

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

XTV_IsPointInContent proc pDispInfo:Pointer, dPosX:dword, dPosY:dword
    mov ecx, [esp + 04]                                 ;pDispInfo
    assume ecx:ptr XTVN_DISP_INFO
    mov edx, [esp + 08]                                 ;dPosX
    xor eax, eax
    .if (sdword ptr edx >= [ecx].ContentRect.left) && (sdword ptr edx < [ecx].ContentRect.right)
      mov edx, [esp + 12]                               ;dPosY
      .if (sdword ptr edx >= [ecx].ContentRect.top) && (sdword ptr edx < [ecx].ContentRect.bottom)
        inc eax
      .endif
    .endif
    assume ecx:NOTHING
    ret 12
XTV_IsPointInContent endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

; 
; Procedure: XTV_IsPointInNode
; Purpose:   Checks if a point is contained in the node band.
; Arguments: Arg1: -> XTVN_DISP_INFO structure.
;            Arg2: y position in client coords.
; Return:    TRUE / FALSE.

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

XTV_IsPointInNode proc pDispInfo:Pointer, dPosY:dword
    mov ecx, [esp + 04]                                 ;pDispInfo
    assume ecx:ptr XTVN_DISP_INFO
    mov edx, [esp + 08]                                 ;dPosY
    xor eax, eax
    .if (sdword ptr edx >= [ecx].sdNodeTop) && (sdword ptr edx < [ecx].sdNodeBot)
      inc eax
    .endif
    assume ecx:NOTHING
    ret 8
XTV_IsPointInNode endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

; 
; Procedure: XTV_IsPointInText
; Purpose:   Checks if a point is contained in the TextRect.
; Arguments: Arg1: -> XTVN_DISP_INFO structure.
;            Arg2: x position in client coords.
;            Arg3: y position in client coords.
; Return:    TRUE / FALSE.

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

XTV_IsPointInText proc pDispInfo:Pointer, dPosX:dword, dPosY:dword
    mov ecx, [esp + 04]                                 ;pDispInfo
    assume ecx:ptr XTVN_DISP_INFO
    mov edx, [esp + 08]                                 ;dPosX
    xor eax, eax
    .if (sdword ptr edx >= [ecx].TextRect.left) && (sdword ptr edx < [ecx].TextRect.right)
      mov edx, [esp + 12]                               ;dPosY
      .if (sdword ptr edx >= [ecx].TextRect.top) && (sdword ptr edx < [ecx].TextRect.bottom)
        inc eax
      .endif
    .endif
    assume ecx:NOTHING
    ret 12
XTV_IsPointInText endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

; 
; Method:    XTreeView.CanDropHere
; Purpose:   Informs how the node on arg1 can be dropped. Std behaviour.
; Arguments: Arg1: -> Node.
;            Arg2: Key flags.
;            Arg3: Cursor position.
; Return:    0 = drop not allowed (XTVN_INSERT_NOT_ALLOWED).
;            1 = drop as first child of reference node (XTVN_INSERT_AS_FIRST_CHILD).
;            2 = drop as last child of reference node (XTVN_INSERT_AS_LAST_CHILD).
;            3 = drop before reference sibling node (XTVN_INSERT_AS_PREV_SIBLING).
;            4 = drop behind reference sibling node (XTVN_INSERT_AS_NEXT_SIBLING).

Method XTreeView.CanDropHere,, pNode:Pointer, dKeyFlags:dword, dCursorPos:dword
    local CursorPos:POINT

    SetObject ecx
    PntS2Pnt CursorPos, dCursorPos
    mov edx, CursorPos.y
    mov eax, CursorPos.x
    ;Check if the cursor is inside the client rect
    .if (sdword ptr edx >= 0) && \                          ;Check first for y coords
        (sdword ptr edx < [ecx].ClientSize.y) && \
        (sdword ptr eax >= 0) && \                          ;and then for x coords
        (sdword ptr eax < [ecx].ClientSize.x)
      mov edx, [ecx].ClientSize.y
      shr edx, 1
      mov eax, pNode
      .if sdword ptr edx > CursorPos.y
        .if eax != [ecx].pFirstChild
          mov eax, XTVN_INSERT_AS_PREV_SIBLING
        .elseif eax != [ecx].pLastChild
          mov eax, XTVN_INSERT_AS_NEXT_SIBLING
        .else
          mov eax, XTVN_INSERT_NOT_ALLOWED
        .endif
      .else
        .if eax != [ecx].pLastChild
          mov eax, XTVN_INSERT_AS_NEXT_SIBLING
        .elseif eax != [ecx].pFirstChild
          mov eax, XTVN_INSERT_AS_PREV_SIBLING
        .else
          mov eax, XTVN_INSERT_NOT_ALLOWED
        .endif
      .endif
    .else
      mov eax, XTVN_INSERT_NOT_ALLOWED
    .endif
MethodEnd, eax

; 
; Method:    XTreeView.Dispatch
; Purpose:   Dispatching of window messages.
; Arguments: Arg1: Window handle.
;            Arg2: Message identifier.
;            Arg3: First message parameter.
;            Arg4: Second message parameter.
; Return:    This value is the result of the message processing and depends on the message.

Method XTreeView.Dispatch, uses esi, hWnd:Handle, uMsg:dword, wParam:dword, lParam:dword
;    DbgMessage uMsg
    SetObject esi
    DispatchEvent <DefWindowProc, hWnd>                 ;esi -> Object instance
MethodEnd, eax

; 
; Method:    XTreeView.Dispose
; Purpose:   Removes the node from the XTreeView and destroyes it with all its childs.
; Arguments: Arg1: -> Node to dispose
; Return:    Nothing.

Method XTreeView.Dispose, uses esi, pNode:Pointer
    SetObject esi
    .if pNode != NULL
      OCall esi::XTreeView.Remove, pNode
      Kill pNode
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.Done
; Purpose:   Finalizes the XTreeView object
; Arguments: None.
; Return:    Nothing.

Method XTreeView.Done, uses esi
    ;DbgText "XTreeView.Done"
    SetObject esi

    ;Destroy all root nodes; child nodes are automatically destroyed!
    mov eax, [esi].pFirstChild
    .while eax != NULL
      push [eax].XTreeViewNode.pNextSibling
      Kill eax
      pop eax
    .endw

    ACall esi.Done
MethodEnd, NONE

; 
; Method:    XTreeView.Draw
; Purpose:   Renders all nodes in view on the back buffer.
; Arguments: None.
; Return:    Nothing.

Method XTreeView.Draw, uses ebx edi esi
    local Rct:RECT, ScrollInfo:SCROLLINFO, RemRect:RECT

    ;DbgText "XTreeView.Draw"
    SetObject esi
    .ifBitClr [esi].dState, XTVS_LOCKED

      ;Some initialization
      OCall [esi].DispInfoColl::DwordCollection.DisposeAll      ;Release all XTVN_DISP_INFO structs

      ;(De)acticate the scrollbars
      xor ecx, ecx
      mov eax, [esi].TotalSize.y
      sub eax, [esi].ClientSize.y
      setns cl
      invoke ShowScrollBar, [esi].hWnd, SB_VERT, ecx
      invoke GetClientRect, [esi].hWnd, addr Rct
      xor ecx, ecx
      mov eax, Rct.right
      sub eax, [esi].TotalSize.x
      sets cl
      invoke ShowScrollBar, [esi].hWnd, SB_HORZ, ecx
      invoke GetClientRect, [esi].hWnd, addr Rct
      xor ecx, ecx
      mov eax, [esi].TotalSize.y
      sub eax, Rct.bottom
      setns cl
      invoke ShowScrollBar, [esi].hWnd, SB_VERT, ecx
      invoke GetClientRect, [esi].hWnd, addr Rct
      m2m [esi].ClientSize.x, Rct.right, eax
      m2m [esi].ClientSize.y, Rct.bottom, eax

      ;Manage the vertical scrollbar
      mov eax, [esi].TotalSize.y
      .if eax > [esi].ClientSize.y
        mov ScrollInfo.SCROLLINFO.cbSize, sizeof SCROLLINFO
        mov ScrollInfo.SCROLLINFO.fMask, SIF_POS or SIF_PAGE

        ;Scale the values to fit in a 16 bit range
        xor edx, edx
        mov eax, [esi].TotalOffset.y
        mov ecx, 0FFFFh
        mul ecx
        div [esi].TotalSize.y
        mov ScrollInfo.SCROLLINFO.nPos, eax

        mov eax, [esi].ClientSize.y
        mov ecx, 0FFFFh
        mul ecx
        div [esi].TotalSize.y
        .if eax == 0
          inc eax
        .endif
        mov ScrollInfo.SCROLLINFO.nPage, eax

        invoke SetScrollInfo, [esi].hWnd, SB_VERT, addr ScrollInfo, TRUE
      .endif

      ;Manage the horizontal scrollbar
      mov eax, [esi].TotalSize.x
      .if eax > [esi].ClientSize.x
        mov ScrollInfo.SCROLLINFO.cbSize, sizeof SCROLLINFO
        mov ScrollInfo.SCROLLINFO.fMask, SIF_POS or SIF_PAGE

        ;Scale the values to fit in a 16 bit range
        xor edx, edx
        mov eax, [esi].TotalOffset.x
        mov ecx, 0FFFFh
        mul ecx
        div [esi].TotalSize.x
        mov ScrollInfo.SCROLLINFO.nPos, eax

        mov eax, [esi].ClientSize.x
        mov ecx, 0FFFFh
        mul ecx
        div [esi].TotalSize.x
        .if eax == 0
          inc eax
        .endif
        mov ScrollInfo.SCROLLINFO.nPage, eax

        invoke SetScrollInfo, [esi].hWnd, SB_HORZ, addr ScrollInfo, TRUE
      .endif

      ;Draw all nodes in view
      .if [esi].pFirstVisibleNode == NULL
        OCall esi.Set1stVisibleNode, [esi].pFirstChild
        mov [esi].TotalOffset.x, 0
        mov [esi].TotalOffset.y, 0
      .endif

      mov edi, [esi].dFirstOffset
      neg edi
      mov ebx, [esi].pFirstVisibleNode
      .if ebx != NULL
        .repeat
          OCall [esi].DispInfoPool::DataPool.NewItem
          mov [eax].XTVN_DISP_INFO.pNode, ebx
          mov [eax].XTVN_DISP_INFO.sdNodeTop, edi
          add edi, [ebx].XTreeViewNode.dHeight
          mov [eax].XTVN_DISP_INFO.sdNodeBot, edi
          OCall [esi].DispInfoColl::DwordCollection.Insert, eax
          OCall ebx::XTreeViewNode.Render, eax
          XTV_BrowseFrwd ebx
          .break .if ebx == NULL
        .until edi >= [esi].ClientSize.y
      .endif

      .if edi < [esi].ClientSize.y
        ;Fill the bottom remaining background
        mov RemRect.left, 0
        m2m RemRect.top, edi
        m2m RemRect.right, [esi].ClientSize.x
        m2m RemRect.bottom, [esi].ClientSize.y
        OCall esi.RenderBackground, addr RemRect
        invoke InvalidateRect, [esi].hWnd, addr RemRect, FALSE
      .endif
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.Drop
; Purpose:   Drops a node.
; Arguments: Arg1: -> Node to drop.
;            Arg2: -> Reference node.
;            Arg3: How to insert when dropped:
;                    - XTVN_INSERT_AS_FIRST_CHILD
;                    - XTVN_INSERT_AS_LAST_CHILD
;                    - XTVN_INSERT_AS_PREV_SIBLING
;                    - XTVN_INSERT_AS_NEXT_SIBLING
; Return:    Nothing.

Method XTreeView.Drop, uses esi, pNode:Pointer, pRefNode:Pointer, dHowToInsert:dword
    SetObject esi
    mov eax, pNode
    .if (eax != NULL) && (eax != pRefNode)

      OCall esi.Remove, pNode
      OCall esi.Insert, pNode, pRefNode, dHowToInsert

      .ifBitSet [esi].dFlags, XTVF_EXPAND_PARENT_ON_DROP
        OCall esi.Select, pNode
        mov eax, pRefNode
        .if (eax != NULL) && (XTV_IsCollapsed(eax))
          OCall eax::XTreeViewNode.Toggle
        .endif
      .else
        OCall esi.Select, pRefNode
      .endif

      OCall esi.Draw
      OCall esi.GetDispInfo, pNode
      .if eax == NULL                                   ;Not in view
        OCall esi.MoveInViewBottom, pNode
        OCall esi.Draw
      .else
        mov edx, [eax].XTVN_DISP_INFO.sdNodeBot
        .if edx >= [esi].ClientSize.y                   ;Only partially visible
          OCall esi.MoveInViewBottom, pNode
          OCall esi.Draw
        .endif
      .endif
      invoke UpdateWindow, [esi].hWnd
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.Find1stVisibleNode
; Purpose:   Searches for the first visible node in view, starting from the current one.
; Arguments: Arg1: new offset
; Return:    Nothing.

Method XTreeView.Find1stVisibleNode, uses ebx esi, dNewOffset:dword
    SetObject esi
    mov eax, dNewOffset
    mov ecx, [esi].pFirstVisibleNode
    mov ebx, [esi].TotalOffset.y
    sub ebx, [esi].dFirstOffset                         ;Check for the top coord of the node

    .if eax < [esi].TotalOffset.y
      ;We are going up
      .while ebx > dNewOffset
        ;Go for next node
        .if [ecx].XTreeViewNode.pPrevSibling != NULL
          mov ecx, [ecx].XTreeViewNode.pPrevSibling
          .if XTV_IsExpanded(ecx)
            sub ebx, [ecx].XTreeViewNode.dBranchHeight
            .if ebx < dNewOffset
              add ebx, [ecx].XTreeViewNode.dBranchHeight
              mov ecx, [ecx].XTreeViewNode.pLastChild
              sub ebx, [ecx].XTreeViewNode.dHeight
              .continue
            .endif
          .endif
          sub ebx, [ecx].XTreeViewNode.dHeight
        .else
          ;Search the upper branch
          mov eax, ecx
          .while TRUE
            .if [eax].XTreeViewNode.pParent == NULL
              jmp @@Found                               ;Stop here
            .else
              mov ecx, [eax].XTreeViewNode.pParent
              sub ebx, [ecx].XTreeViewNode.dHeight
              .break
            .endif
          .endw
        .endif
      .endw

    .else
      ;We are going down
      add ebx, [ecx].XTreeViewNode.dHeight              ;Check for the bottom coord of the node

      .while ebx < dNewOffset
        .if XTV_IsExpanded(ecx)
          add ebx, [ecx].XTreeViewNode.dBranchHeight
          .if ebx > dNewOffset
            sub ebx, [ecx].XTreeViewNode.dBranchHeight
            mov ecx, [ecx].XTreeViewNode.pFirstChild
            add ebx, [ecx].XTreeViewNode.dHeight
            .continue
          .endif
        .endif

        ;Go for next node
        .if [ecx].XTreeViewNode.pNextSibling != NULL
          mov ecx, [ecx].XTreeViewNode.pNextSibling
          add ebx, [ecx].XTreeViewNode.dHeight
        .else
          ;Search the upper branch
          mov eax, ecx
          .while TRUE
            .if [eax].XTreeViewNode.pParent == NULL
              jmp @F                                    ;Stop here
            .else
              mov eax, [eax].XTreeViewNode.pParent
              .if [eax].XTreeViewNode.pNextSibling != NULL
                mov ecx, [eax].XTreeViewNode.pNextSibling
                add ebx, [ecx].XTreeViewNode.dHeight
                .break
              .endif
            .endif
          .endw
        .endif
      .endw
@@:
      sub ebx, [ecx].XTreeViewNode.dHeight
    .endif

@@Found:
    mov eax, dNewOffset
    mov [esi].TotalOffset.y, eax
    sub eax, ebx
    mov [esi].dFirstOffset, eax
    OCall esi.Set1stVisibleNode, ecx

MethodEnd, NONE

; 
; Method:    XTreeView.GetDispInfo
; Purpose:   Returns the XTVN_DISP_INFO structure associated to a XTreeViewNode that is in view.
; Arguments: Arg1: -> XTreeViewNode
; Return:    eax -> XTVN_DISP_INFO, NULL if not found.
;            ecx = index in DispInfoColl.

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

XTV_IsNodeInDispData proc pDispData:Pointer, pNode:Pointer, pIndex:Pointer
    mov eax, [esp + 04]                                 ;pDispData
    mov ecx, [esp + 08]                                 ;pNode
    .if ecx == [eax].XTVN_DISP_INFO.pNode
      mov eax, TRUE
    .else
      mov edx, [esp + 12]                               ;pIndex
      inc dword ptr [edx]
      xor eax, eax
    .endif
    ret 12
XTV_IsNodeInDispData endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

Method XTreeView.GetDispInfo, NOFRAME, pNode:Pointer
    SetObject ecx,, [esp + 4]
    sub esp, 4
    mov dword ptr [esp], 0
    OCall [ecx].DispInfoColl::DwordCollection.FirstThat, \
                              offset XTV_IsNodeInDispData, Pointer ptr [esp + 16], esp    ;pNode
    .if eax == NULL
      mov ecx, -1
      add esp, 4
    .else
      pop ecx
    .endif
MethodEnd 8, eax ecx

; 
; Method:    XTreeView.Init
; Purpose:   Initializes a XTreeView object.
; Arguments: Arg1: -> owner object.
;            Arg2: Parent handle.
;            Arg2: -> Definition structure.
; Return:    Nothing.

Method XTreeView.Init, uses esi, pOwner:Pointer, hParentWnd:Handle, pDefStruc:Pointer
    SetObject esi
    mov ecx, pDefStruc
    assume ecx:ptr XTV_DEF
    mov eax, [ecx].dStyle
    or eax, WS_CHILD or WS_VISIBLE or WS_TABSTOP
    invoke CreateWindowEx, [ecx].dExStyle, offset szXTV_Class, [ecx].pTitle, \
           eax, [ecx].sdPosX, [ecx].sdPosY, [ecx].dWidth, [ecx].dHeight, \
           hParentWnd, [ecx].dCtlID, hInstance, esi
    assume ecx:NOTHING
    ACall esi.Init, pOwner, [esi].hWnd

MethodEnd, NONE

; 
; Method:    XTreeView.InitRes
; Purpose:   Initializes the XTreeViewNode object from a resource.
; Arguments: Window Handle.
; Return:    Nothing.

Method XTreeView.InitRes, uses esi, hWnd:Handle
    SetObject esi
    BitSet [esi].dFlags, XTVF_RESOURCE
    m2m [esi].hWnd, hWnd
    ACall esi.Init, NULL, hWnd
MethodEnd, NONE

; 
; Method:    XTreeView.Insert
; Purpose:   Inserts a new node.
; Arguments: Arg1: -> Node to insert.
;            Arg2: -> Reference node. NULL is considdered as the root.
;            Arg3: Insertion mode. See "XTreeViewNode.CanDropHere" method for valid values.
; Return:    eax -> inserted node.

Method XTreeView.Insert, uses ebx esi, pNode:Pointer, pRefNode:Pointer, dHowToInsert:dword
    SetObject esi
    mov eax, pNode
    .if eax != NULL

      ;Set the correct parent node
      .if (dHowToInsert == XTVN_INSERT_AS_FIRST_CHILD) || \
          (dHowToInsert == XTVN_INSERT_AS_LAST_CHILD)
        m2m [eax].XTreeViewNode.pParent, pRefNode
      .elseif (dHowToInsert == XTVN_INSERT_AS_PREV_SIBLING) || \
              (dHowToInsert == XTVN_INSERT_AS_NEXT_SIBLING)
        mov ecx, pRefNode
        .if ecx == NULL
          push ecx
        .else
          push [ecx].XTreeViewNode.pParent
        .endif
        pop [eax].XTreeViewNode.pParent
      .else
        OCall esi.ErrorReport, NULL, XTV_UNKNOWN_INSERTION_MODE
        xor eax, eax
        ExitMethod
      .endif

      ;Adjust tree metrics
      mov eax, pNode
      mov edx, [eax].XTreeViewNode.dHeight
      mov ebx, [eax].XTreeViewNode.dWidth
      .if XTV_IsExpanded(eax)
        add edx, [eax].XTreeViewNode.dBranchHeight          ;Cummulate heights
        mov ecx, [eax].XTreeViewNode.dBranchWidth
        add ecx, [esi].dNodeIndent
        Maxi ebx, ecx, eax                                  ;Trash ecx and eax, ebx = max width
        mov eax, pNode
      .endif
      mov ecx, [eax].XTreeViewNode.pParent
      .if (ecx != NULL) && ([ecx].XTreeViewNode.dState & XTVNS_EXPANDED)
        .repeat
          add [ecx].XTreeViewNode.dBranchHeight, edx        ;Cummulate heights
          push edx
          mov edx, ebx
          Maxi [ecx].XTreeViewNode.dBranchWidth, edx, eax   ;Trash edx and eax
          mov edx, [ecx].XTreeViewNode.dWidth
          add ebx, [esi].dNodeIndent
          Maxi ebx, edx, eax                                ;Trash edx and eax
          pop edx
          mov ecx, [ecx].XTreeViewNode.pParent              ;Move to parent node
        .until (ecx == NULL) || !([ecx].XTreeViewNode.dState & XTVNS_EXPANDED)
        .if ecx == NULL                                     ;Did we reach the root node?
          add [esi].TotalSize.y, edx
          add ebx, [esi].dNodeIndent
          add ebx, [esi].ClientOffset.x
          Maxi [esi].TotalSize.x, ebx, eax                  ;Trash ebx and eax
        .endif
      .else
        .if ecx != NULL
          add [ecx].XTreeViewNode.dBranchHeight, edx
          add ebx, [esi].dNodeIndent
          Maxi [ecx].XTreeViewNode.dBranchWidth, ebx, eax   ;Trash ebx and eax
        .else
          add [esi].TotalSize.y, edx
          add ebx, [esi].dNodeIndent
          add ebx, [esi].ClientOffset.x
          Maxi [esi].TotalSize.x, ebx, eax                  ;Trash ebx and eax
        .endif
      .endif

      ;Insert the node
      .if dHowToInsert == XTVN_INSERT_AS_FIRST_CHILD
        mov eax, pNode
        mov [eax].XTreeViewNode.pPrevSibling, NULL
        .if [eax].XTreeViewNode.pParent == NULL
          mov edx, [esi].XTreeView.pFirstChild
          mov [esi].XTreeView.pFirstChild, eax
          mov [eax].XTreeViewNode.pNextSibling, edx
          .if edx == NULL
            mov [esi].XTreeView.pLastChild, eax
          .else
            mov [edx].XTreeViewNode.pPrevSibling, eax
          .endif
        .else
          mov ecx, [eax].XTreeViewNode.pParent
          mov edx, [ecx].XTreeViewNode.pFirstChild
          mov [ecx].XTreeViewNode.pFirstChild, eax
          mov [eax].XTreeViewNode.pNextSibling, edx
          .if edx == NULL
            mov [ecx].XTreeViewNode.pLastChild, eax
          .else
            mov [edx].XTreeViewNode.pPrevSibling, eax
          .endif
        .endif

      .elseif dHowToInsert == XTVN_INSERT_AS_LAST_CHILD
        mov eax, pNode
        mov [eax].XTreeViewNode.pNextSibling, NULL
        .if [eax].XTreeViewNode.pParent == NULL
          mov edx, [esi].XTreeView.pLastChild
          mov [eax].XTreeViewNode.pPrevSibling, edx
          mov [esi].XTreeView.pLastChild, eax
          .if edx == NULL
            mov [esi].XTreeView.pFirstChild, eax
          .else
            mov [edx].XTreeViewNode.pNextSibling, eax
          .endif
        .else
          mov ecx, [eax].XTreeViewNode.pParent
          mov edx, [ecx].XTreeViewNode.pLastChild
          mov [ecx].XTreeViewNode.pLastChild, eax
          mov [eax].XTreeViewNode.pPrevSibling, edx
          .if edx == NULL
            mov [ecx].XTreeViewNode.pFirstChild, eax
          .else
            mov [edx].XTreeViewNode.pNextSibling, eax
          .endif
        .endif

      .elseif dHowToInsert == XTVN_INSERT_AS_PREV_SIBLING
        mov eax, pNode
        .if [eax].XTreeViewNode.pParent == NULL
          mov edx, pRefNode
          .if edx == NULL
            mov edx, [esi].pFirstChild
            mov [esi].XTreeView.pFirstChild, eax
            mov [eax].XTreeViewNode.pNextSibling, edx
            mov [eax].XTreeViewNode.pPrevSibling, NULL
            .if edx == NULL
              mov [esi].XTreeView.pLastChild, eax
            .else
              mov [edx].XTreeViewNode.pPrevSibling, eax
            .endif
          .else
            mov ecx, [edx].XTreeViewNode.pPrevSibling
            mov [eax].XTreeViewNode.pNextSibling, edx
            mov [eax].XTreeViewNode.pPrevSibling, ecx
            mov [edx].XTreeViewNode.pPrevSibling, eax
            .if ecx == NULL
              mov [esi].pFirstChild, eax
            .else
              mov [ecx].XTreeViewNode.pNextSibling, eax
            .endif
          .endif
        .else
          mov edx, pRefNode
          mov ecx, [edx].XTreeViewNode.pPrevSibling
          mov [eax].XTreeViewNode.pNextSibling, edx
          mov [edx].XTreeViewNode.pPrevSibling, eax
          mov [eax].XTreeViewNode.pPrevSibling, ecx
          .if ecx == NULL
            mov ecx, [edx].XTreeViewNode.pParent
            mov [ecx].XTreeViewNode.pFirstChild, eax
          .else
            mov [ecx].XTreeViewNode.pNextSibling, eax
          .endif
        .endif

      .elseif dHowToInsert == XTVN_INSERT_AS_NEXT_SIBLING
        mov eax, pNode
        .if [eax].XTreeViewNode.pParent == NULL
          mov edx, pRefNode
          .if edx == NULL
            mov edx, [esi].pLastChild
            mov [esi].XTreeView.pLastChild, eax
            mov [eax].XTreeViewNode.pPrevSibling, edx
            mov [eax].XTreeViewNode.pNextSibling, NULL
            .if edx == NULL
              mov [esi].XTreeView.pFirstChild, eax
            .else
              mov [edx].XTreeViewNode.pNextSibling, eax
            .endif
          .else
            mov ecx, [edx].XTreeViewNode.pNextSibling
            mov [eax].XTreeViewNode.pPrevSibling, edx
            mov [eax].XTreeViewNode.pNextSibling, ecx
            mov [edx].XTreeViewNode.pNextSibling, eax
            .if ecx == NULL
              mov [esi].pLastChild, eax
            .else
              mov [ecx].XTreeViewNode.pPrevSibling, eax
            .endif
          .endif
        .else
          mov edx, pRefNode
          mov ecx, [edx].XTreeViewNode.pNextSibling
          mov [edx].XTreeViewNode.pNextSibling, eax
          mov [eax].XTreeViewNode.pNextSibling, ecx
          mov [eax].XTreeViewNode.pPrevSibling, edx
          .if ecx == NULL
            mov ecx, [edx].XTreeViewNode.pParent
            mov [ecx].XTreeViewNode.pLastChild, eax
          .else
            mov [ecx].XTreeViewNode.pPrevSibling, eax
          .endif
        .endif
      .endif

    .else
      OCall esi.ErrorReport, NULL, XTV_INVALID_POINTER
      xor eax, eax
    .endif
MethodEnd, eax

; 
; Method:    XTreeView.MoveInViewBottom
; Purpose:   Brings the specified node to the bottom of the current view.
; Arguments: Arg1: -> Node to be moved.
; Return:    Nothing.
; Notes:     - no visual update is performed!

Method XTreeView.MoveInViewBottom, uses ebx esi, pNode:Pointer
    SetObject esi

    .if pNode != NULL
      ;Reset hottrack flag
      mov eax, [esi].pHotTrackedNode
      .if eax != NULL
        BitClr [eax].XTreeViewNode.dState, XTVNS_HOTTRACK
      .endif

      mov ecx, pNode
      mov ebx, [ecx].XTreeViewNode.dHeight

      .while ebx < [esi].ClientSize.y
        XTV_BrowseBack ecx
        .break .if ecx == NULL
        add ebx, [ecx].XTreeViewNode.dHeight
      .endw

      .if ecx == NULL
        mov [esi].TotalOffset.y, ecx
        mov [esi].dFirstOffset, ecx
        OCall esi.Set1stVisibleNode, [esi].pFirstChild
      .else
        sub ebx, [esi].ClientSize.y
        mov [esi].dFirstOffset, ebx
        push ecx
        OCall esi.Set1stVisibleNode, ecx
        pop ecx
        .while TRUE
          .if [ecx].XTreeViewNode.pPrevSibling != NULL
            mov ecx, [ecx].XTreeViewNode.pPrevSibling
            .if XTV_IsExpanded(ecx)
              add ebx, [ecx].XTreeViewNode.dBranchHeight
            .endif
            add ebx, [ecx].XTreeViewNode.dHeight
          .else
            mov ecx, [ecx].XTreeViewNode.pParent
            .break .if ecx == NULL
            add ebx, [ecx].XTreeViewNode.dHeight
          .endif
        .endw
        mov [esi].TotalOffset.y, ebx
      .endif

      invoke InvalidateRect, [esi].hWnd, NULL, FALSE
    .else
      OCall esi.ErrorReport, NULL, XTV_INVALID_POINTER
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.MoveInViewTop
; Purpose:   Brings the specified node on the top of the current view.
; Arguments: Arg1: -> Node.
; Return:    Nothing.
; Notes:     - no visual update is performed!

Method XTreeView.MoveInViewTop, uses esi, pNode:Pointer
    SetObject esi

    .if pNode != NULL
      ;Reset hottrack flag
      mov eax, [esi].pHotTrackedNode
      .if eax != NULL
        BitClr [eax].XTreeViewNode.dState, XTVNS_HOTTRACK
      .endif

      xor edx, edx
      mov ecx, pNode

      .while TRUE
        .if [ecx].XTreeViewNode.pPrevSibling != NULL
          mov ecx, [ecx].XTreeViewNode.pPrevSibling
          .if XTV_IsExpanded(ecx)
            add edx, [ecx].XTreeViewNode.dBranchHeight
          .endif
          add edx, [ecx].XTreeViewNode.dHeight
        .else
          mov ecx, [ecx].XTreeViewNode.pParent
          .break .if ecx == NULL
          add edx, [ecx].XTreeViewNode.dHeight
        .endif
      .endw

      mov [esi].TotalOffset.y, edx
      OCall esi.Set1stVisibleNode, pNode
      mov [esi].dFirstOffset, 0

      invoke InvalidateRect, [esi].hWnd, NULL, FALSE

    .else
      OCall esi.ErrorReport, NULL, XTV_INVALID_POINTER

    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.OnButtonDblClk
; Purpose:   Handler of mouse double clicks.
; Arguments: Arg1: Button that triggered this method.
;            Arg2: First message parameter. Key flags.
;            Arg3: Second message parameter. Cursor position.
; Return:    Nothing.

Method XTreeView.OnButtonDblClk, uses esi, dButton:dword, wParam:dword, lParam:dword
    local CursorPos:POINT

    ;DbgText "XTreeView.OnButtonDblClk"
    SetObject esi
    PntS2Pnt CursorPos, lParam
    OCall [esi].DispInfoColl::DwordCollection.FirstThat, addr XTV_IsPointInNode, CursorPos.y
    .if eax != NULL
      lea ecx, CursorPos
      OCall [eax].XTVN_DISP_INFO.pNode::XTreeViewNode.MouseHitDbl, dButton, ecx, eax, wParam
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.OnButtonDown
; Purpose:   Handler of mouse button down.
; Arguments: Arg1: Button that triggered this method.
;            Arg2: First message parameter. Key flags.
;            Arg3: Second message parameter. Cursor position.
; Return:    Nothing.

Method XTreeView.OnButtonDown, uses esi, dButton:dword, wParam:dword, lParam:dword
    local CursorPos:POINT

    SetObject esi
    invoke SetFocus, [esi].hWnd

    PntS2Pnt CursorPos, lParam
    OCall [esi].DispInfoColl::DwordCollection.FirstThat, addr XTV_IsPointInNode, CursorPos.y
    .if eax != NULL
      lea ecx, CursorPos
      OCall [eax].XTVN_DISP_INFO.pNode::XTreeViewNode.MouseHit, dButton, ecx, eax, wParam
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.OnButtonUp
; Purpose:   Handler of mouse button up.
; Arguments: Arg1: Button that triggered this method.
;            Arg2: First message parameter. Key flags.
;            Arg3: Second message parameter. Cursor position.
; Return:    Nothing.

Method XTreeView.OnButtonUp, uses esi, dButton:dword, wParam:dword, lParam:dword
    SetObject esi
    BitClr [esi].dState, XTVS_DRAGGINGSTARTED
    mov eax, [esi].dDragMouseButton
    .if (eax == dButton) && ([esi].dState & XTVS_DRAGGING)
      invoke ReleaseCapture
      BitClr [esi].dState, XTVS_DRAGGING
      invoke SetCursor, [esi].hCursorArrow

      ;Remove the selected node and reinsert it using the hottracked node as insertion reference
      .if [esi].pHotTrackedNode == NULL
        OCall esi.CanDropHere, [esi].pSelectedNode, wParam, lParam
      .else
        OCall [esi].pHotTrackedNode::XTreeViewNode.CanDropHere, [esi].pSelectedNode, wParam, lParam
      .endif
      .if eax != 0
        OCall esi.Drop, [esi].pSelectedNode, [esi].pHotTrackedNode, eax
      .endif

    .elseIfBitSet [esi].dState, XTVS_DRAGGINGABORTED
      BitClr [esi].dState, XTVS_DRAGGINGABORTED

    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.OnCreate
; Purpose:   Event handler for WM_CREATE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    eax = Zero to continue creation of the window, else -1.

Method XTreeView.OnCreate, uses esi, wParam:dword, lParam:dword
    local hDC:Handle, ScrollInfo:SCROLLINFO

    ;DbgText "XTreeView.OnCreate"
    SetObject esi

    ;Some initialization
    mov [esi].hCursorArrow, $invoke(LoadCursor, hInstance, "CURSOR_ARROW")
    mov [esi].hCursorNo,    $invoke(LoadCursor, hInstance, "CURSOR_NO")
    mov [esi].hCursorNode,  $invoke(LoadCursor, hInstance, "CURSOR_ARROWNODE")
    mov [esi].hCursorFirst, $invoke(LoadCursor, hInstance, "CURSOR_ARROWFIRST")
    mov [esi].hCursorLast,  $invoke(LoadCursor, hInstance, "CURSOR_ARROWLAST")
    mov [esi].hCursorPrev,  $invoke(LoadCursor, hInstance, "CURSOR_ARROWPREV")
    mov [esi].hCursorNext,  $invoke(LoadCursor, hInstance, "CURSOR_ARROWNEXT")
    OCall [esi].DispInfoColl::DwordCollection.Init, esi, 20, 10, COLLECTIONMAXSIZE
    OCall [esi].DispInfoPool::DataPool.Init, esi, [esi].dDispInfoSize, 4096
    lea ecx, [esi].DispInfoColl
    Override ecx::DwordCollection.DestroyItem, DisplayInfoColl.DestroyItem

    ;Create the text font
    mov hDC, $invoke(GetDC, [esi].hWnd)
    invoke GetDeviceCaps, hDC, LOGPIXELSY
    invoke MulDiv, 10, eax, -72
    mov XTV_LogFontText.lfHeight, eax
    mov [esi].hFontText, $invoke(CreateFontIndirect, offset XTV_LogFontText)
    invoke ReleaseDC, [esi].hWnd, hDC

    ;Create pens to draw
    mov [esi].hPenLine, $invoke(CreatePen, PS_SOLID, 0, [esi].dColorLine)
    mov [esi].hPenToggler, $invoke(CreatePen, PS_SOLID, 0, [esi].dColorToggler)

    ;Set scroller infos
    mov ScrollInfo.cbSize, sizeof SCROLLINFO
    mov ScrollInfo.fMask, SIF_RANGE
    mov ScrollInfo.nMin, 0
    mov ScrollInfo.nMax, 0FFFFh       ;16 bit max value passed in WM_HSCROLL/WM_VSCROLL
    invoke SetScrollInfo, [esi].hWnd, SB_VERT, addr ScrollInfo, FALSE
    invoke SetScrollInfo, [esi].hWnd, SB_HORZ, addr ScrollInfo, FALSE

    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnEnable
; Purpose:   Event handler for WM_ENABLE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnEnable, uses esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnEnable"
    SetObject esi
    mov eax, wParam
    mov edx, [esi].dState
    and edx, XTVS_DISABLED
    xor edx, eax
    .if Zero?                                           ;Has the enable status changed?
      .if eax != FALSE
        BitClr [esi].dState, XTVS_DISABLED
      .else
        BitSet [esi].dState, XTVS_DISABLED
      .endif
      OCall esi.Draw
      invoke UpdateWindow, [esi].hWnd
    .endif
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnDestroy
; Purpose:   Event handler for WM_DESTROY message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnDestroy, uses esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnDestroy"

    SetObject esi

    OCall [esi].DispInfoColl::DwordCollection.Done
    OCall [esi].DispInfoPool::DataPool.Done

    invoke DeleteObject, [esi].hPenLine
    invoke DeleteObject, [esi].hPenToggler
    invoke DeleteObject, [esi].hFontText

    invoke DeleteObject, $invoke(SelectObject, [esi].hMemDC, [esi].hBmpPrv)
    invoke DeleteDC, [esi].hMemDC

    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnGetDlgCode
; Purpose:   Event handler for WM_GETDLGCODE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    eax indicates which type of input the control processes.

Method XTreeView.OnGetDlgCode,, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnGetDlgCode"
    mov eax, DLGC_WANTARROWS or DLGC_WANTCHARS
MethodEnd, eax

; 
; Method:    XTreeView.OnGetFont
; Purpose:   Event handler for WM_GETFONT message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    eax = the font with which the control is currently drawing its text.

Method XTreeView.OnGetFont,, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnGetFont"
    SetObject ecx
    mov eax, [ecx].hFontText
MethodEnd, eax

; 
; Method:    XTreeView.OnHScroll
; Purpose:   Event handler for WM_HSCROLL message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XTreeView.OnHScroll, uses esi, wParam:dword, lParam:dword
    SetObject esi
    mov eax, wParam
    .if ax == SB_THUMBPOSITION || ax == SB_THUMBTRACK
      shr eax, 16                                       ;Get high part of eax
      mov ecx, [esi].TotalSize.x
      mul ecx
      mov ecx, 0FFFFh
      div ecx
      mov ecx, [esi].TotalSize.x
      sub ecx, [esi].ClientSize.x
      .if eax > ecx
        mov eax, ecx
      .endif

    .elseif ax == SB_LINEUP
      mov edx, [esi].ArrowScrollValue.x
      .if edx < [esi].TotalOffset.x
        mov eax, [esi].TotalOffset.x
        sub eax, edx
        inc [esi].ArrowScrollValue.x
      .else
        xor eax, eax
      .endif

    .elseif ax == SB_LINEDOWN
      mov edx, [esi].ArrowScrollValue.x
      mov eax, [esi].TotalSize.x
      sub eax, [esi].ClientSize.x
      mov ecx, eax
      sub ecx, edx
      .if sdword ptr ecx > [esi].TotalOffset.x
        mov eax, [esi].TotalOffset.x
        add eax, edx
        inc [esi].ArrowScrollValue.x
      .endif

    .elseif ax == SB_PAGEUP
      mov edx, [esi].ClientSize.x
      .if edx < [esi].TotalOffset.x
        mov eax, [esi].TotalOffset.x
        sub eax, edx
      .else
        xor eax, eax
      .endif

    .elseif ax == SB_PAGEDOWN
      mov edx, [esi].ClientSize.x
      mov eax, [esi].TotalSize.x
      sub eax, edx
      mov ecx, eax
      sub ecx, edx
      .if sdword ptr ecx > [esi].TotalOffset.x
        mov eax, [esi].TotalOffset.x
        add eax, edx
      .endif

    .else
      jmp @@Exit                                        ;Unrecognized message
    .endif
    mov [esi].TotalOffset.x, eax

    OCall esi.Draw
    invoke UpdateWindow, [esi].hWnd
@@Exit:
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnKeyDown
; Purpose:   Event handler for WM_KEYDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XTreeView.OnKeyDown, uses ebx esi, wParam:dword, lParam:dword
    local CRect:RECT, CursorPos:POINT

    SetObject esi
    .if [esi].pSelectedNode != NULL
      mov eax, wParam

      .if eax == VK_UP
        ;Move to the above node
        mov ecx, [esi].pSelectedNode
        mov ebx, ecx
        XTV_BrowseBack ecx
        cmp ecx, NULL
        jz @@Exit                                       ;Nothing to do, avoid flicker
        OCall esi.Select, ecx
        OCall esi.GetDispInfo, [esi].pSelectedNode
        .if ecx == 0                                    ;First node in view
          mov [esi].dFirstOffset, 0
          mov eax, [esi].pFirstChild
          .if eax == [esi].pSelectedNode
            mov [esi].TotalOffset.y, 0
          .endif
          OCall esi.Draw
        .elseif eax == NULL                             ;Not in view
          OCall esi.MoveInViewTop, [esi].pSelectedNode
          OCall esi.Draw
        .else
          OCall [esi].pSelectedNode::XTreeViewNode.RenderContent, eax
          OCall esi.GetDispInfo, ebx
          OCall ebx::XTreeViewNode.RenderContent, eax
        .endif

      .elseif eax == VK_DOWN
        ;Move to the below node
        mov ecx, [esi].pSelectedNode
        mov ebx, ecx
        XTV_BrowseFrwd ecx
        cmp ecx, NULL
        jz @@Exit                                       ;Nothing to do, avoid flicker
        OCall esi.Select, ecx
        OCall esi.GetDispInfo, [esi].pSelectedNode
        mov edx, [esi].DispInfoColl.dCount
        dec edx
        .if ecx >= edx                                  ;Is it the last node in view,
          OCall esi.MoveInViewBottom, [esi].pSelectedNode   ;  or is it not in view
          OCall esi.Draw
        .else
          OCall [esi].pSelectedNode::XTreeViewNode.RenderContent, eax
          OCall esi.GetDispInfo, ebx
          OCall ebx::XTreeViewNode.RenderContent, eax
        .endif

      .elseif eax == VK_RIGHT
        ;First expand if collapsed and then move to the first child
        mov ebx, [esi].pSelectedNode
        cmp [ebx].XTreeViewNode.pFirstChild, NULL
        jz @@Exit                                       ;Nothing to do, avoid flicker
        .ifBitClr [ebx].XTreeViewNode.dState, XTVNS_EXPANDED
          OCall ebx::XTreeViewNode.Toggle
        .else
          mov ebx, [ebx].XTreeViewNode.pFirstChild
          OCall esi.Select, ebx
        .endif

        ;Node is now expanded; maybe we have to reposition the tree
        OCall esi.GetDispInfo, ebx
        mov eax, [eax].XTVN_DISP_INFO.sdNodeTop
        add eax, [esi].TotalOffset.y                    ;eax = node offset
        mov edx, eax
        add edx, [ebx].XTreeViewNode.dHeight
        add edx, [ebx].XTreeViewNode.dBranchHeight
        sub edx, [esi].ClientSize.y
        sub edx, [esi].TotalOffset.y
        .if !Sign?
          mov ecx, eax                                  ;The last branch node is off view
          sub ecx, [esi].TotalOffset.y
          sub ecx, edx
          .if !Sign?
            add eax, [ebx].XTreeViewNode.dHeight
            add eax, [ebx].XTreeViewNode.dBranchHeight
            sub eax, [esi].ClientSize.y
          .endif
          OCall esi.Find1stVisibleNode, eax
        .endif
        OCall esi.Draw

      .elseif eax == VK_LEFT
        ;First collapse if expanded and then move to the parent
        mov ecx, [esi].pSelectedNode
        .if XTV_IsExpanded(ecx)
          OCall ecx::XTreeViewNode.Toggle

          ;Correct the client area coords
          mov eax, [esi].TotalSize.x
          .if eax > [esi].ClientSize.x
            invoke ShowScrollBar, [esi].hWnd, SB_HORZ, TRUE
            invoke GetClientRect, [esi].hWnd, addr CRect
            m2m [esi].ClientSize.x, CRect.right
            m2m [esi].ClientSize.y, CRect.bottom
          .endif

          XTV_GetBottomViewNode esi, ecx, eax, edx
          .if ecx == NULL
            OCall esi.MoveInViewBottom, edx
          .endif
        .else
          cmp [ecx].XTreeViewNode.pParent, NULL
          jz @@Exit                                     ;Nothing to do, avoid flicker
          mov ecx, [ecx].XTreeViewNode.pParent
          push ecx
          OCall esi.Select, ecx
          pop ecx
          OCall esi.GetDispInfo, ecx
          .if ecx == 0                                  ;First node in view
            mov [esi].dFirstOffset, 0
          .elseif eax == NULL                           ;Not in view
            OCall esi.MoveInViewTop, [esi].pSelectedNode
          .endif
        .endif
        ;Check if the selected node is in view
        XTV_GetBottomViewNode esi, ecx, eax, edx
        .if ecx == NULL
          OCall esi.MoveInViewBottom, edx
        .endif
        OCall esi.Draw

      .elseif eax == VK_PGUP
        mov eax, [esi].pFirstChild
        cmp eax, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker

        ;Move the selected node on bottom of the view
        OCall esi.MoveInViewBottom, [esi].pSelectedNode

        ;Get the first completely visible node and select it
        mov ecx, [esi].pSelectedNode
        mov edx, [ecx].XTreeViewNode.dHeight
        .repeat
          mov eax, ecx
          XTV_BrowseBack ecx
          .break .if ecx == NULL
          add edx, [ecx].XTreeViewNode.dHeight
        .until edx > [esi].ClientSize.y

        ;Select it and adjust the view if it is the first node
        OCall esi.Select, eax
        XTV_GetBottomViewNode esi, ecx, edx, eax
        .if ecx == NULL
          OCall esi.MoveInViewBottom, eax
        .else
          OCall esi.MoveInViewTop, [esi].pSelectedNode
        .endif
        OCall esi.Draw

      .elseif eax == VK_PGDN
        mov eax, [esi].pLastChild
        cmp eax, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker

        ;Move the selected node on top of the view
        OCall esi.MoveInViewTop, [esi].pSelectedNode

        ;Get the last completely visible node and select it
        mov ecx, [esi].pSelectedNode
        mov edx, [ecx].XTreeViewNode.dHeight
        .repeat
          mov eax, ecx
          XTV_BrowseFrwd ecx
          .break .if ecx == NULL
          add edx, [ecx].XTreeViewNode.dHeight
        .until edx > [esi].ClientSize.y

        ;Select it and adjust the view if it is the last node
        push eax
        push ecx
        OCall esi.Select, eax
        pop ecx
        pop eax
        .if ecx == NULL
          OCall esi.MoveInViewBottom, eax
        .endif
        OCall esi.Draw

      .elseif eax == VK_DELETE
        OCall esi.Remove, [esi].pSelectedNode
        .if [esi].pSelectedNode == NULL
          ;All nodes deleted
          xor eax, eax
          mov [esi].TotalOffset.y, eax
          mov [esi].TotalOffset.x, eax
          mov [esi].dFirstOffset, eax
          OCall esi.Set1stVisibleNode, eax
        .else
          .if eax == [esi].pFirstVisibleNode            ;The deleted node was the first in view
            OCall esi.MoveInViewTop, [esi].pSelectedNode
          .else
            mov edx, [esi].TotalSize.y
            sub edx, [esi].ClientSize.y
            .if sdword ptr edx < [esi].TotalOffset.y    ;Have we at hole on the bottom of the view?
              .if sdword ptr edx < 0
                xor edx, edx
              .endif
              OCall esi.Find1stVisibleNode, edx

              ;Check if the selected node is in view; if not, bring it in top position
              mov edx, [esi].pSelectedNode
              mov ecx, [esi].pFirstVisibleNode
              .while (ecx != NULL) && (ecx != edx)
                XTV_BrowseFrwd ecx
              .endw
              .if ecx == NULL
                OCall esi.MoveInViewTop, edx
              .endif

            .endif
          .endif
        .endif
        OCall esi.Draw

      .elseif eax == VK_HOME
        ;Go to the first node in the tree
        mov ecx, [esi].pFirstChild
        cmp ecx, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker
        push ecx
        OCall esi.Set1stVisibleNode, ecx
        pop ecx
        OCall esi.Select, ecx
        mov [esi].TotalOffset.y, 0
        mov [esi].dFirstOffset, 0
        OCall esi.Draw

      .elseif eax == VK_END
        ;Go to the last node in the tree
        mov ecx, [esi].pLastChild
        .while ecx != NULL
          .break .if XTV_IsCollapsed(ecx)
          mov edx, ecx
          mov ecx, [ecx].XTreeViewNode.pLastChild
        .endw
        .if ecx == NULL
          mov ecx, edx
        .endif
        cmp ecx, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker
        push ecx
        OCall esi.Select, ecx
        pop ecx
        OCall esi.MoveInViewBottom, ecx
        OCall esi.Draw

      .elseif eax == VK_ESCAPE
        BitClr [esi].dState, XTVS_DRAGGINGSTARTED
        .ifBitSet [esi].dState, XTVS_DRAGGING
          invoke ReleaseCapture
          BitClr [esi].dState, XTVS_DRAGGING
          BitSet [esi].dState, XTVS_DRAGGINGABORTED
          invoke SetCursor, [esi].hCursorArrow
        .endif

      .elseif eax == VK_CONTROL
        XTV_SimulateMouseMove CONTROL, LBUTTON, MBUTTON, RBUTTON, SHIFT


      .elseif eax == VK_SHIFT
        XTV_SimulateMouseMove SHIFT, LBUTTON, MBUTTON, RBUTTON, CONTROL

      .else
        jmp @@Exit
      .endif
      invoke UpdateWindow, [esi].hWnd
      xor eax, eax

    .else
@@Exit:
      invoke DefWindowProc, [esi].hWnd, WM_KEYDOWN, wParam, lParam

    .endif
MethodEnd

; 
; Method:    XTreeView.OnKeyUp
; Purpose:   Event handler for WM_KEYUP message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XTreeView.OnKeyUp, uses esi, wParam:dword, lParam:dword
    local CursorPos:POINT

    SetObject esi
    .if [esi].pSelectedNode != NULL
      mov eax, wParam

      .if eax == VK_CONTROL
        XTV_SimulateMouseMove 0, LBUTTON, MBUTTON, RBUTTON, SHIFT

      .elseif eax == VK_SHIFT
        XTV_SimulateMouseMove 0, LBUTTON, MBUTTON, RBUTTON, CONTROL

      .else
        jmp @@Exit
      .endif
      invoke UpdateWindow, [esi].hWnd
      xor eax, eax

    .else
@@Exit:
      invoke DefWindowProc, [esi].hWnd, WM_KEYUP, wParam, lParam

    .endif
MethodEnd, eax

; 
; Method:    XTreeView.OnKillFocus
; Purpose:   Event handler for WM_KILLFOCUS message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.
; Notes:     - Only affects the selected and hottracked nodes.
;              Redefine this method if other processing is required.

Method XTreeView.OnKillFocus, uses ebx esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnKillFocus"
    SetObject esi
    BitClr [esi].dState, (XTVS_FOCUSED or XTVS_DRAGGINGSTARTED)
    .ifBitSet [esi].dState, XTVS_DRAGGING
      invoke ReleaseCapture
      BitClr [esi].dState, XTVS_DRAGGING
      BitSet [esi].dState, XTVS_DRAGGINGABORTED
      invoke SetCursor, [esi].hCursorArrow
    .endif

    xor ebx, ebx

    .if [esi].pSelectedNode != NULL
      OCall esi.GetDispInfo, [esi].pSelectedNode
      OCall [esi].pSelectedNode::XTreeViewNode.RenderContent, eax
      inc ebx
    .endif

    mov eax, [esi].pHotTrackedNode
    .if eax != NULL
      BitClr [eax].XTreeViewNode.dState, XTVNS_HOTTRACK
      OCall esi.GetDispInfo, eax
      OCall [esi].pHotTrackedNode::XTreeViewNode.RenderContent, eax
      inc ebx
    .endif

    .if ebx != 0
      invoke UpdateWindow, [esi].hWnd
    .endif

    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnLButtonDblClk
; Purpose:   Event handler for WM_LBUTTONDBLCLK message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnLButtonDblClk, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonDblClk, MK_LBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnLButtonDown
; Purpose:   Event handler for WM_LBUTTONDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnLButtonDown, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonDown, MK_LBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnLButtonUp
; Purpose:   Event handler for for WM_LBUTTONUP message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnLButtonUp, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonUp, MK_LBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnMButtonDblClk
; Purpose:   Event handler for WM_MBUTTONDBLCLK message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnMButtonDblClk, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonDblClk, MK_MBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnMButtonDown
; Purpose:   Event handler for WM_MBUTTONDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnMButtonDown, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonDown, MK_MBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnMButtonUp
; Purpose:   Event handler for for WM_MBUTTONUP message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnMButtonUp, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonUp, MK_MBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnMouseMove
; Purpose:   Event handler for WM_MOUSEMOVE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

XTV_EmitMouseMoveMsg proc hWnd:Handle, wParam:dword
    local CursorPos:POINT

    invoke GetCursorPos, addr CursorPos                     ;Get current (new) cursor position
    invoke ScreenToClient, hWnd, addr CursorPos
    mov eax, CursorPos.y
    shl eax, 16
    mov ax, word ptr CursorPos.x
    invoke PostMessage, hWnd, WM_MOUSEMOVE, wParam, eax     ;Retrigger WM_MOUSEMOVE
    ret
XTV_EmitMouseMoveMsg endp

Method XTreeView.OnMouseMove, uses ebx esi, wParam:dword, lParam:dword
    local CRect:RECT, CursorPos:POINT, pDispInfo:Pointer

    SetObject esi
    ;DbgText "XTreeView.OnMouseMove"

    .if !([esi].dState & XTVS_DISABLED) && ([esi].dState & XTVS_FOCUSED)
      invoke SetCapture, [esi].hWnd
      PntS2Pnt CursorPos, lParam
      xor ebx, ebx
      invoke GetClientRect, [esi].hWnd, addr CRect
      invoke IsPntInRect,  addr CursorPos, addr CRect
      .if eax != FALSE
        OCall [esi].DispInfoColl::DwordCollection.FirstThat, addr XTV_IsPointInText, CursorPos
        .if eax != NULL
          mov ebx, [eax].XTVN_DISP_INFO.pNode
        .endif
        mov eax, wParam
        and eax, MK_LBUTTON or MK_MBUTTON or MK_RBUTTON
        .if (eax == [esi].dDragMouseButton) && (ebx != [esi].pSelectedNode) && \
            ([esi].dState & XTVS_DRAGGINGSTARTED) && !([esi].dState & XTVS_DRAGGING)
          BitClr [esi].dState, XTVS_DRAGGINGSTARTED
          BitSet [esi].dState, XTVS_DRAGGING
          invoke SetCursor, [esi].hCursorNode
          mov ecx, [esi].pSelectedNode
          .if (ecx != NULL) && XTV_IsExpanded(ecx) && ([esi].dFlags & XTVF_COLLAPSE_ON_DRAG)
            OCall ecx::XTreeViewNode.Toggle
            XTV_GetBottomViewNode esi, edx, eax
            .if edx == NULL
              OCall esi.MoveInViewBottom, [esi].pSelectedNode
            .endif
            OCall esi.Draw
          .endif
          jmp @@Exit
        .endif

        ;Check if the user want to scroll
        .ifBitSet [esi].dState, XTVS_DRAGGING
          mov eax, [esi].dScrollActivation
          .if sdword ptr eax > CursorPos.y
            mov eax, [esi].TotalOffset.y
            .if sdword ptr eax > 0
              dec eax
              OCall esi.Find1stVisibleNode, eax
              OCall esi.Draw
              invoke XTV_EmitMouseMoveMsg, [esi].hWnd, wParam
            .endif
          .else
            mov eax, [esi].ClientSize.y
            sub eax, [esi].dScrollActivation
            .if sdword ptr eax <= CursorPos.y
              mov ecx, [esi].TotalSize.y
              sub ecx, [esi].ClientSize.y
              .if sdword ptr ecx > [esi].TotalOffset.y
                mov eax, [esi].TotalOffset.y
                inc eax
                OCall esi.Find1stVisibleNode, eax
                OCall esi.Draw
                invoke XTV_EmitMouseMoveMsg, [esi].hWnd, wParam
              .endif
            .else
              mov eax, [esi].dScrollActivation
              .if sdword ptr eax > CursorPos.x
                mov eax, [esi].TotalOffset.x
                .if sdword ptr eax > 0
                  dec [esi].TotalOffset.x
                  OCall esi.Draw
                  invoke XTV_EmitMouseMoveMsg, [esi].hWnd, wParam
                .endif
              .else
                mov eax, [esi].ClientSize.x
                sub eax, [esi].dScrollActivation
                .if sdword ptr eax <= CursorPos.x
                  mov ecx, [esi].TotalSize.x
                  sub ecx, [esi].ClientSize.x
                  .if sdword ptr ecx > [esi].TotalOffset.x
                    inc [esi].TotalOffset.x
                    OCall esi.Draw
                    invoke XTV_EmitMouseMoveMsg, [esi].hWnd, wParam
                  .endif
                .endif
              .endif
            .endif
          .endif
        .endif

        ;Hottrack management
        OCall [esi].DispInfoColl::DwordCollection.FirstThat, addr XTV_IsPointInContent, CursorPos
        .if eax != NULL
          mov pDispInfo, eax
          mov ebx, [eax].XTVN_DISP_INFO.pNode
          .if ebx != [esi].pHotTrackedNode
            BitSet [ebx].XTreeViewNode.dState, XTVNS_HOTTRACK
            OCall ebx::XTreeViewNode.RenderContent, pDispInfo
            xchg ebx, [esi].pHotTrackedNode
            .if ebx != NULL
              BitClr [ebx].XTreeViewNode.dState, XTVNS_HOTTRACK
              OCall esi.GetDispInfo, ebx
              OCall ebx::XTreeViewNode.RenderContent, eax
            .endif
          .endif
        .else
          .if [esi].pHotTrackedNode != NULL
            mov ebx, [esi].pHotTrackedNode
            BitClr [ebx].XTreeViewNode.dState, XTVNS_HOTTRACK
            OCall esi.GetDispInfo, ebx
            OCall ebx::XTreeViewNode.RenderContent, eax
            mov [esi].pHotTrackedNode, NULL
          .endif
        .endif

        .ifBitSet [esi].dState, XTVS_DRAGGING
          OCall esi.SetDragCursor, wParam, lParam
        .else
          invoke SetCursor, [esi].hCursorArrow
        .endif

      .else
        .ifBitSet [esi].dState, XTVS_DRAGGING
          invoke SetCursor, [esi].hCursorNode           ;In case we have hCursorNo
        .else
          invoke ReleaseCapture
        .endif
        mov ebx, [esi].pHotTrackedNode
        .if ebx != NULL
          BitClr [ebx].XTreeViewNode.dState, XTVNS_HOTTRACK
          OCall esi.GetDispInfo, ebx
          OCall ebx::XTreeViewNode.RenderContent, eax
        .endif
        mov [esi].pHotTrackedNode, NULL

      .endif

@@Exit:
      invoke UpdateWindow, [esi].hWnd
   .endif

    xor eax, eax

MethodEnd, eax

; 
; Method:    XTreeView.OnMouseWheel
; Purpose:   Event handler for WM_MOUSEWHEEL message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.
; Note:      if the shift key is pressed, the horizontal scroller is moved.

Method XTreeView.OnMouseWheel, uses esi, wParam:dword, lParam:dword
    local CursorPos:POINT

    ;DbgText "XTreeView.OnMouseWheel"
    SetObject esi
    mov eax, wParam
    xor edx, edx
    sar eax, 16
    .if !Zero?                                          ;Exit if no increment
      .if SIGN?                                         ;Is eax negative?
        dec edx                                         ;Sign extend edx for idiv
      .endif
      mov ecx, 12                                       ;/120 * 10
      idiv ecx
      neg eax                                           ;Change sign to toggle scroll direction

      .if word ptr [wParam] == MK_SHIFT                 ;Is shift key pressed?
        push [esi].TotalOffset.x
        mov ecx, [esi].TotalSize.x
        sub ecx, [esi].ClientSize.x
        js @@Exit                                       ;Nothing to draw, avoid flicker
        add eax, [esi].TotalOffset.x
        .if SIGN?                                       ;Is it negative?
          xor eax, eax                                  ;No negative values allowed!
        .else
          Mini eax, ecx, edx                            ;Limit to the upper value
        .endif
        pop ecx
        .if ecx != eax
          mov [esi].TotalOffset.x, eax
        .else
          jmp @@Exit                                    ;Nothing to draw, avoid flicker
        .endif

      .else
        push [esi].TotalOffset.y
        mov ecx, [esi].TotalSize.y
        sub ecx, [esi].ClientSize.y
        js @@Exit                                       ;Nothing to draw, avoid flicker
        add eax, [esi].TotalOffset.y
        .if SIGN?
          xor eax, eax
        .else
          Mini eax, ecx, edx                            ;Limit to the upper value
        .endif

        OCall esi.Find1stVisibleNode, eax
        pop eax
        .if eax == [esi].TotalOffset.y
          jmp @@Exit                                    ;Nothing to draw, avoid flicker
        .endif
      .endif

      OCall esi.Draw

      ;Manage the hottracked node
      mov eax, [esi].pHotTrackedNode
      .if eax != NULL
        OCall esi.GetDispInfo, eax
        .if eax != NULL
          mov ecx, [eax].XTVN_DISP_INFO.pNode
          BitClr [ecx].XTreeViewNode.dState, XTVNS_HOTTRACK
          mov [esi].pHotTrackedNode, NULL
          OCall ecx::XTreeViewNode.RenderContent, eax
        .endif
      .endif

      invoke GetCursorPos, addr CursorPos
      invoke ScreenToClient, [esi].hWnd, addr CursorPos
      OCall [esi].DispInfoColl::DwordCollection.FirstThat, addr XTV_IsPointInContent, CursorPos
      .if eax != NULL
        mov ecx, [eax].XTVN_DISP_INFO.pNode
        BitSet [ecx].XTreeViewNode.dState, XTVNS_HOTTRACK
        mov [esi].pHotTrackedNode, ecx
        OCall ecx::XTreeViewNode.RenderContent, eax
      .else
        mov [esi].pHotTrackedNode, eax
      .endif

      invoke UpdateWindow, [esi].hWnd
@@Exit:
      xor eax, eax
    .endif
MethodEnd, eax

; 
; Method:    XTreeView.OnNcDestroy
; Purpose:   Event handler for WM_NCDESTROY message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnNcDestroy, uses esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnNcDestroy"
    SetObject esi
    .ifBitSet [esi].dFlags, XTVF_RESOURCE
      invoke SetWindowLong, [esi].hWnd, GWL_USERDATA, NULL
      Destroy esi                                       ;Selfdestruction only if it was created
    .endif                                              ;  from a resource
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnNcHitTest
; Purpose:   Event handler for WM_NCHITTEST message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    eax = DefWindowProc return value.

Method XTreeView.OnNcHitTest, NOFRAME, wParam:dword, lParam:dword
    SetObject ecx,, [esp + 4]
    mov [ecx].ArrowScrollValue.x, 1
    mov [ecx].ArrowScrollValue.y, 1
    invoke DefWindowProc, [ecx].hWnd, WM_NCHITTEST, [esp + 12], [esp + 12]  ;wParam, lParam
MethodEnd 12, eax

; 
; Method:    XTreeView.OnPaint
; Purpose:   Event handler for WM_PAINT message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnPaint, uses esi, wParam:dword, lParam:dword
    local PS:PAINTSTRUCT

    ;DbgText "XTreeView.OnPaint"
    SetObject esi
    .ifBitSet [esi].dState, XTVS_VISIBLE
      .if $invoke(GetUpdateRect, [esi].hWnd, NULL, FALSE) != 0
        invoke BeginPaint, [esi].hWnd, addr PS
        mov eax, PS.rcPaint.right
        sub eax, PS.rcPaint.left
        mov ecx, PS.rcPaint.bottom
        sub ecx, PS.rcPaint.top
        invoke BitBlt, PS.hdc, PS.rcPaint.left, PS.rcPaint.top, eax, ecx, \
                               [esi].hMemDC, PS.rcPaint.left, PS.rcPaint.top, SRCCOPY
        invoke EndPaint, [esi].hWnd, addr PS
      .endif
    .endif
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnRButtonDblClk
; Purpose:   Event handler for WM_RBUTTONDBLCLK message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnRButtonDblClk, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonDblClk, MK_RBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnRButtonDown
; Purpose:   Event handler for WM_RBUTTONDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnRButtonDown, NOFRAME, wParam:dword, lParam:dword
    OCall Pointer ptr [esp + 16]::XTreeView.OnButtonDown, MK_RBUTTON, \
            dword ptr [esp + 12], dword ptr [esp + 12]  ;wParam, lParam
    xor eax, eax
MethodEnd 12, eax

; 
; Method:    XTreeView.OnRButtonUp
; Purpose:   Event handler for WM_RBUTTONUP message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnRButtonUp, uses esi, wParam:dword, lParam:dword
    local CursorPos:POINT

    SetObject esi
    OCall esi.OnButtonUp, MK_RBUTTON, wParam, lParam

    .ifBitClr [esi].dState, XTVS_DRAGGING
      PntS2Pnt CursorPos, lParam
      OCall [esi].DispInfoColl::DwordCollection.FirstThat, addr XTV_IsPointInText, CursorPos
      .if eax != NULL
        push eax
        invoke ClientToScreen, [esi].hWnd, addr CursorPos
        pop eax
        OCall [eax].XTVN_DISP_INFO.pNode::XTreeViewNode.ContextPopup, CursorPos
      .endif
    .endif

    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnSize
; Purpose:   Event handler for WM_SIZE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnSize, uses esi, wParam:dword, lParam:dword
    local hDC:Handle

    ;DbgText "XTreeView.OnSize"
    SetObject esi

    .if [esi].hMemDC != 0
      invoke DeleteObject, $invoke(SelectObject, [esi].hMemDC, [esi].hBmpPrv)
      invoke DeleteDC, [esi].hMemDC
    .endif

    mov hDC, $invoke(GetDC, [esi].hWnd)
    mov [esi].hMemDC, $invoke(CreateCompatibleDC, hDC)
    PntS2Pnt [esi].ClientSize, lParam
    mov [esi].hBmpMem, $invoke(CreateCompatibleBitmap, hDC, [esi].ClientSize.x, [esi].ClientSize.y)
    mov [esi].hBmpPrv, $invoke(SelectObject, [esi].hMemDC, [esi].hBmpMem)
    invoke ReleaseDC, [esi].hWnd, hDC

    ;Check if TotalOffset.y has still a valid value
    mov eax, [esi].TotalSize.y
    sub eax, [esi].ClientSize.y
    .if Sign?
      OCall esi.Set1stVisibleNode, [esi].pFirstChild    ;Move the first node on top of the view
      mov [esi].TotalOffset.y, 0
      mov [esi].dFirstOffset, 0
    .elseif eax < [esi].TotalOffset.y
      OCall esi.Find1stVisibleNode, eax                 ;Reposition the window content
    .endif

    ;Check if TotalOffset.x has still a valid value
    mov eax, [esi].TotalSize.x
    sub eax, [esi].ClientSize.x
    .if Sign?
      mov [esi].TotalOffset.x, 0
    .elseif eax < [esi].TotalOffset.x
      mov [esi].TotalOffset.x, eax                      ;Reposition the window content
    .endif

    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnSetFocus
; Purpose:   Event handler for WM_SETFOCUS message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnSetFocus, uses esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnSetFocus"
    SetObject esi
    BitSet [esi].dState, XTVS_FOCUSED
    .if [esi].pSelectedNode != NULL
      OCall esi.GetDispInfo, [esi].pSelectedNode
      OCall [esi].pSelectedNode::XTreeViewNode.RenderContent, eax
      invoke UpdateWindow, [esi].hWnd
    .endif
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnSetFont
; Purpose:   Event handler for WM_SETFONT message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Nothing.

Method XTreeView.OnSetFont, uses esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnSetFont"
    SetObject esi
    .if [esi].hFontText != 0
      invoke DeleteObject, [esi].hFontText
      m2m [esi].hFontText, wParam
    .endif
    .if lParam != 0
      OCall esi.Draw
      invoke UpdateWindow, [esi].hWnd
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.OnSetRedraw
; Purpose:   Event handler for WM_SETREDRAW message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XTreeView.OnSetRedraw,, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnSetRedraw"
    SetObject ecx
    .if wParam == FALSE
      BitSet [ecx].dState, XTVS_LOCKED
    .else
      BitClr [ecx].dState, XTVS_LOCKED
    .endif
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.OnVScroll
; Purpose:   Event handler for WM_VSCROLL message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XTreeView.OnVScroll, uses esi, wParam:dword, lParam:dword
    ;DbgText "XTreeView.OnVScroll"
    SetObject esi
    mov eax, wParam
    .if ax == SB_THUMBPOSITION || ax == SB_THUMBTRACK
      shr eax, 16                                       ;Get high part of eax
      mov ecx, [esi].TotalSize.y
      mul ecx
      mov ecx, 0FFFFh
      div ecx
      mov ecx, [esi].TotalSize.y
      sub ecx, [esi].ClientSize.y
      .if eax > ecx
        mov eax, ecx
      .endif

    .elseif ax == SB_LINEUP
      mov edx, [esi].ArrowScrollValue.y
      .if edx < [esi].TotalOffset.y
        mov eax, [esi].TotalOffset.y
        sub eax, edx
        inc [esi].ArrowScrollValue.y
      .else
        xor eax, eax
      .endif

    .elseif ax == SB_LINEDOWN
      mov edx, [esi].ArrowScrollValue.y
      mov eax, [esi].TotalSize.y
      sub eax, [esi].ClientSize.y
      mov ecx, eax
      sub ecx, edx
      .if sdword ptr ecx > [esi].TotalOffset.y
        mov eax, [esi].TotalOffset.y
        add eax, edx
        inc [esi].ArrowScrollValue.y
      .endif

    .elseif ax == SB_PAGEUP
      mov edx, [esi].ClientSize.y
      .if edx < [esi].TotalOffset.y
        mov eax, [esi].TotalOffset.y
        sub eax, edx
      .else
        xor eax, eax
      .endif

    .elseif ax == SB_PAGEDOWN
      mov edx, [esi].ClientSize.y
      mov eax, [esi].TotalSize.y
      sub eax, edx
      mov ecx, eax
      sub ecx, edx
      .if sdword ptr ecx > [esi].TotalOffset.y
        mov eax, [esi].TotalOffset.y
        add eax, edx
      .endif

    .else
      jmp @@Exit                                        ;Unrecognized message
    .endif

    OCall esi.Find1stVisibleNode, eax
    OCall esi.Draw
    invoke UpdateWindow, [esi].hWnd
@@Exit:
    xor eax, eax
MethodEnd, eax

; 
; Method:    XTreeView.Remove
; Purpose:   Removes an inserted node from the XTreeiew.
; Arguments: Arg1: -> to the node to remove.
; Return:    eax -> removed node.

Method XTreeView.Remove, uses ebx edi esi, pNode:Pointer
    SetObject esi
    .if pNode != NULL

      ;Manage the selected node if it is in the deleted branch
      mov ecx, pNode
      .ifBitSet [ecx].XTreeViewNode.dState, XTVNS_SEL_PATH
        .if [ecx].XTreeViewNode.pNextSibling != NULL
          mov eax, [ecx].XTreeViewNode.pNextSibling
        .elseif [ecx].XTreeViewNode.pPrevSibling != NULL
          mov eax, [ecx].XTreeViewNode.pPrevSibling
        .else
          mov eax, [ecx].XTreeViewNode.pParent
        .endif
        OCall esi.Select, eax
      .endif

      ;Manage the first node in view
      mov ecx, pNode
      .ifBitSet [ecx].XTreeViewNode.dState, XTVNS_VIS_PATH
        .if [ecx].XTreeViewNode.pPrevSibling != NULL
          mov eax, [ecx].XTreeViewNode.pPrevSibling
        .elseif [ecx].XTreeViewNode.pParent != NULL
          mov eax, [ecx].XTreeViewNode.pParent
        .else
          mov eax, [ecx].XTreeViewNode.pNextSibling
        .endif
        OCall esi.Set1stVisibleNode, eax
      .endif

      ;Remove the node
      mov eax, pNode
      .if [eax].XTreeViewNode.pParent == NULL
        mov edx, [eax].XTreeViewNode.pPrevSibling
        .if edx != NULL
          m2m [edx].XTreeViewNode.pNextSibling, [eax].XTreeViewNode.pNextSibling
        .else
          m2m [esi].XTreeView.pFirstChild, [eax].XTreeViewNode.pNextSibling
        .endif
        mov edx, [eax].XTreeViewNode.pNextSibling
        .if edx != NULL
          m2m [edx].XTreeViewNode.pPrevSibling, [eax].XTreeViewNode.pPrevSibling
        .else
          m2m [esi].XTreeView.pLastChild, [eax].XTreeViewNode.pPrevSibling
        .endif

      .else
        mov ecx, [eax].XTreeViewNode.pParent
        mov edx, [eax].XTreeViewNode.pPrevSibling
        .if edx != NULL
          m2m [edx].XTreeViewNode.pNextSibling, [eax].XTreeViewNode.pNextSibling
        .else
          m2m [ecx].XTreeViewNode.pFirstChild, [eax].XTreeViewNode.pNextSibling
        .endif
        mov edx, [eax].XTreeViewNode.pNextSibling
        .if edx != NULL
          m2m [edx].XTreeViewNode.pPrevSibling, [eax].XTreeViewNode.pPrevSibling
        .else
          m2m [ecx].XTreeViewNode.pLastChild, [eax].XTreeViewNode.pPrevSibling
        .endif

      .endif
      ;eax -> deleted node

      ;Calc new branch values in the modified path
      mov edi, [eax].XTreeViewNode.pParent
      mov ecx, [eax].XTreeViewNode.dHeight
      .if XTV_IsExpanded(eax)
        add ecx, [eax].XTreeViewNode.dBranchHeight
      .endif

      push ecx
      .while (edi != NULL) && ([edi].XTreeViewNode.dState & XTVNS_EXPANDED)
        mov ecx, dword ptr [esp]
        sub [edi].XTreeViewNode.dBranchHeight, ecx
        XTV_GetBranchMaxWidth [edi].XTreeViewNode.pFirstChild, esi
        mov [edi].XTreeViewNode.dBranchWidth, ebx
        mov edi, [edi].XTreeViewNode.pParent            ;Move one level up
      .endw
      pop ecx

      .if edi == NULL                                   ;Did we reach the root node?
        sub [esi].TotalSize.y, ecx

        XTV_GetBranchMaxWidth [esi].pFirstChild, esi
        add ebx, [esi].dNodeIndent
        add ebx, [esi].ClientOffset.x
        mov [esi].TotalSize.x, ebx
      .endif

      mov eax, pNode                                    ;Return value

    .else
      OCall esi.ErrorReport, NULL, XTV_INVALID_POINTER
      xor eax, eax
    .endif
MethodEnd, eax

; 
; Method:    XTreeView.RenderBackground
; Purpose:   Draws the background of a given rect.
; Arguments: Arg1: -> RECT.
; Return:    Nothing.

Method XTreeView.RenderBackground, uses esi, pRect:Pointer
    SetObject esi
    invoke CreateSolidBrush, [esi].dColorBackground
    push eax
    push eax
    push pRect
    push [esi].hMemDC
    call FillRect
    call DeleteObject
MethodEnd, NONE

; 
; Method:    XTreeView.Select
; Purpose:   Selects a new node.
; Arguments: Arg1: -> node to be selected. NULL is allowed.
; Return:    Nothing.

Method XTreeView.Select, NOFRAME, pNode:Pointer
    ;DbgText "XTreeView.Select"
    SetObject eax,, [esp + 4]
    mov ecx, [esp + 8]                                  ;pNode
    mov edx, [eax].pSelectedNode
    .if ecx != edx
      mov [eax].pSelectedNode, ecx
      .while edx != NULL
        BitClr [edx].XTreeViewNode.dState, XTVNS_SEL_PATH
        mov edx, [edx].XTreeViewNode.pParent
      .endw
      .while ecx != NULL
        BitSet [ecx].XTreeViewNode.dState, XTVNS_SEL_PATH
        mov ecx, [ecx].XTreeViewNode.pParent
      .endw
    .endif
MethodEnd 8, NONE

; 
; Method:    XTreeView.Set1stVisibleNode
; Purpose:   Sets the first visible node and marks the path to it, resetting the prevoius.
;            NULL is allowed.
; Arguments: Arg1: -> node to be selected.
; Return:    Nothing.

Method XTreeView.Set1stVisibleNode, NOFRAME, pNode:Pointer
    SetObject eax,, [esp + 4]
    mov ecx, [esp + 8]                                  ;pNode
    mov edx, [eax].pFirstVisibleNode
    mov [eax].pFirstVisibleNode, ecx
    .while edx != NULL
      BitClr [edx].XTreeViewNode.dState, XTVNS_VIS_PATH
      mov edx, [edx].XTreeViewNode.pParent
    .endw
    .while ecx != NULL
      BitSet [ecx].XTreeViewNode.dState, XTVNS_VIS_PATH
      mov ecx, [ecx].XTreeViewNode.pParent
    .endw
MethodEnd 8, NONE

; 
; Method:    XTreeView.SetDragCursor
; Purpose:   While dragging, sets the cursor over a hottracked node.
; Arguments: Arg1: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
;            Arg2: Cursor position.
; Return:    Nothing.

Method XTreeView.SetDragCursor, uses esi, dKeyFlags:dword, CursorPos:dword
    SetObject esi
    mov ecx, [esi].pHotTrackedNode
    .if ecx == [esi].pSelectedNode
      invoke SetCursor, [esi].hCursorNo
    .else
      .if ecx != NULL
        OCall ecx::XTreeViewNode.CanDropHere, [esi].pSelectedNode, dKeyFlags, CursorPos
      .else
        OCall esi.CanDropHere, [esi].pSelectedNode, dKeyFlags, CursorPos
      .endif
      .if eax == XTVN_INSERT_NOT_ALLOWED
        .if [esi].pHotTrackedNode == NULL
          mov edx, [esi].hCursorNode
        .else
          mov edx, [esi].hCursorNo
        .endif
      .elseif eax == XTVN_INSERT_AS_FIRST_CHILD
        mov edx, [esi].hCursorFirst
      .elseif eax == XTVN_INSERT_AS_LAST_CHILD
        mov edx, [esi].hCursorLast
      .elseif eax == XTVN_INSERT_AS_PREV_SIBLING
        mov edx, [esi].hCursorPrev
      .elseif eax == XTVN_INSERT_AS_NEXT_SIBLING
        mov edx, [esi].hCursorNext
      .else
        mov edx, [esi].hCursorNode
      .endif
      invoke SetCursor, edx
    .endif
MethodEnd, NONE

; 
; Method:    XTreeView.Startup
; Purpose:   Registers the XTreeView object with the OS.
; Arguments: None.
; Return:    Nothing.

Method XTreeView.Startup
    local WC:WNDCLASSEX

    mov WC.cbSize, sizeof WNDCLASSEX
    mov WC.style, CS_HREDRAW or CS_VREDRAW or CS_PARENTDC or CS_BYTEALIGNWINDOW or \
                  CS_BYTEALIGNCLIENT; or CS_DBLCLKS
    m2m WC.lpfnWndProc, $MethodAddr(XTreeView.WndProc)
    mov WC.cbClsExtra, 0
    mov WC.cbWndExtra, 0
    m2m WC.hInstance, hInstance
    mov WC.hbrBackground, 0
    mov WC.lpszMenuName, NULL
    mov WC.lpszClassName, offset szXTV_Class
    mov WC.hIcon, 0

    mov WC.hCursor, $invoke(LoadCursor, hInstance, "CURSOR_ARROW")
    mov WC.hIconSm, 0

    invoke RegisterClassEx, addr WC
MethodEnd, NONE

; 
; Method:    XTreeView.WindowProc
; Purpose:   Processing of window messages.
; Arguments: Arg1: Message identifier.
;            Arg2: First message parameter.
;            Arg3: Second message parameter.
;            Window Handle is passed in Self.
; Return:    This value is the result of the message processing and depends on the message.

Method XTreeView.WndProc, uses esi, uMsg:dword, wParam:dword, lParam:dword
    ;DbgMessage uMsg
    .if uMsg == WM_NCCREATE
      mov eax, lParam
      .if [eax].CREATESTRUCT.lpCreateParams == NULL
        ;if lpCreateParams is 0, then it is a resource call => create new object instance
        mov esi, $New(XTreeView, InitRes, Self)
      .else
        ;Init call
        mov esi, [eax].CREATESTRUCT.lpCreateParams
        m2m [esi].XTreeView.hWnd, Self
      .endif
      invoke SetWindowLong, Self, GWL_USERDATA, esi
      OCall esi::XTreeView.Dispatch, Self, uMsg, wParam, lParam
    .else
      invoke GetWindowLong, Self, GWL_USERDATA              ;if GWL_USERDATA wasn't set,
      .if eax == NULL                                       ;GetWindowLong returns NULL
        invoke DefWindowProc, Self, uMsg, wParam, lParam
      .else
        OCall eax::XTreeView.Dispatch, Self, uMsg, wParam, lParam
      .endif
    .endif
MethodEnd, eax

;##/



; ==================================================================================================
;     DispInfoColl
; ==================================================================================================

;##\

; 
; Method:    DisplayInfoColl.DestroyItem
; Purpose:   Auxiliary procedure. Don't call directly!
; Arguments: -> item.
; Return:    Nothing.

Method DisplayInfoColl.DestroyItem, NOFRAME, pItem:Pointer
    SetObject ecx,, [esp + 4]
    mov edx, [ecx].DwordCollection.pOwner
    OCall [edx].XTreeView.DispInfoPool::DataPool.FreeItem, Pointer ptr [esp + 8]    ;pItem
MethodEnd 8, NONE

;##/


endif
